"""
Exception list

    errors:



"""


class AloError(Exception):
    """
    Raised when undefined error.
    Read the error message and fix it.
    If you cannot fix the error, Please contact the administrator with log file.

    정의되지 않은 예외 발생.
    에러 메시지를 읽고, 해당 내용을 수정하세요.
    에러의 원인을 찾거나, 수정할 수 없다면, 로그 파일을 관리자에게 전달 후 조치 방법을 문의하세요.
    """

    codes = {}
    code = None  # error code
    fmt = '[{code}] : {message}'

    def __init__(self, message, exception=None, doc={}, **kwargs):
        try:
            msg = f"{self.fmt.format(code=self.code, message=message, **kwargs)}\n{self.__doc__.format(**{'code':self.code, **doc}) if doc else self.__doc__ }"
            Exception.__init__(self, msg)
        except Exception:
            Exception.__init__(self, message + "\nCheck error format.")

        self.kwargs = kwargs

    def __init_subclass__(cls, **kwargs):
        if not cls.code:
            raise Exception("A constant value must be assigned to the 'code' variable as a class member variable.")
        if AloError.codes.get(cls.code):
            raise Exception(f"{cls.code} is duplicated : {[AloError.codes.get(cls.code).__name__, cls.__name__]}")

        AloError.codes[cls.code] = cls

    @classmethod
    def print(cls):
        print("code,name,description,document")
        print(f"{cls.code},{cls.__name__},{cls.fmt}")
        for k, v in cls.codes.items():
            print(f"{v.__name__},{k},{v.fmt}")


class AloValueError(AloError):
    """
    Raised when found invalid key or value error.
    Check key or value
    """
    code = "ALO-VAL-000"
    fmt = '[{code}] Found invalid value : {message}'


#################
# Alo Init Error
class AloInitError(AloError):
    """
    Raised when function initialization fails.
    Check required field, network, git, redis, experimental plan or solution meta...
    """

    code = "ALO-INI-000"
    fmt = '[{code}] Failed alo init : {message}'


class AloInitGitError(AloError):
    """
    Raised when branch not found in upstream origin.
    Check the same branch name exists in the upstream origin.
    """

    code = "ALO-INI-001"
    fmt = '[{code}] Failed install alolib : {message}'


class AloInitRequirementError(AloError):
    """
    Raised when the 3rd party library cannot be installed.
    Check the 3rd party library list and version in the requirements.txt file..
    """

    code = "ALO-INI-002"
    fmt = '[{code}] Failed installing alolib requirements.txt : {message}'


class AloInitFileNotFountError(AloError):
    """
    Raised when no such file or directory.
    Check file path.
    """

    code = "ALO-INI-003"
    fmt = '[{code}] The file or directory does not exist: {message}'


class AloInitRedisError(AloError):
    """
    Raised when unable to connect to redis.
    Check redis config, or redis status
    """

    code = "ALO-INI-004"
    fmt = '[{code}] Failed to connect redis : {message}'


class AloInitInvalidKeyValueError(AloError):
    """
    Raised when set an invalid key or value.
    Check the key and value values in the yaml(or value).
    """

    code = "ALO-INI-005"
    fmt = '[{code}] Found an invalid keys or values in file : {file}\n{message}'


class AloInitJsonInvalidError(AloError):
    """
    Raised when set an invalid json string.
    Check json string.
    """

    code = "ALO-INI-006"
    fmt = '[{code}] Found an invalid keys or values in json string : {message}'

class AloInitNameError(AloError):

    """
    에러 코드
        {code}

    에러 정보
        - Invalid call: {file}
        - Invalid name: {key}

    에러 원인
        제작하신 AI Soltuion의 {file}이 python 파일 이름 혹은 함수 명에 'alo' 혹은 'main'을 포함합니다.

    조치 가이드
        | {file}의 'alo' 혹은 'main' 부분을 수정해주세요.
        | Ex) titanic.main -> titanic.train # titanic.py 내 main 함수 명 수정 시
        | Ex) main.train -> titanic.train   # main.py 파일 명 수정 시
    """

    code = "ALO-INI-007"

#################
# pipeline Error
class AloPipelineInitError(AloError):
    """
    Raised when initializing the pipeline.
    """

    code = "ALO-PIP-000"
    fmt = '[{code}] Check the yaml file : {message}'


class AloPipelineAssetError(AloError):
    """
    Raised when failed to set up the assets in the scripts folder based on whether the code source is local or git.
    """

    code = "ALO-PIP-001"
    fmt = '[{code}] Check the code source of asset or git : {message}'


class AloPipelineImportError(AloError):
    """
    Raised when failed to import module/function.
    Check the library name and version in the requirements.
    Or Add the library path to the PYTHONPATH environment variable.
    """

    code = "ALO-PIP-002"
    fmt = '[{code}] Failed to import module/function - {module} : {message}'


class AloPipelineBatchError(AloError):
    """
    Raised when failed operate batch job.
    """

    code = "ALO-PIP-003"
    fmt = '[{code}] Check the {pipeline}(pipeline) : {message}'


class AloPipelineArtifactError(AloError):
    """
    Raised when Failed to empty & re-make artifacts.
    """

    code = "ALO-PIP-004"
    fmt = '[{code}] Check pipeline : {pipeline} - {message}'


class AloPipelineRequirementsError(AloError):
    """
    Raised when the 3rd party library cannot be installed.
    Check the library version or network.
    """

    code = "ALO-PIP-005"
    fmt = '[{code}] Found error when installing the package : {pipeline} - {message}'


class AloPipelineBackupError(AloError):
    """
    Raised when backup error history & save error artifacts.
    """

    code = "ALO-PIP-006"
    fmt = '[{code}] Check : {pipeline} - {message}'


class AloPipelineLoadError(AloError):
    """
    Raised when loading external data or duplicated basename in the same pipeline.
    """

    code = "ALO-PIP-007"
    fmt = '[{code}] Failed to load(get) : {pipeline} - {message}'


class AloPipelineSaveError(AloError):
    """
    Raised when save artifacts.
    """

    code = "ALO-PIP-008"
    fmt = '[{code}] Failed to save : {pipeline} - {message}'


class AloPipelineConversionError(AloError):
    """
    Raised when converting data in the execution unit.
    Check solutioin_metadata or experimental_plan.
    """

    code = "ALO-PIP-009"
    fmt = '[{code}] Error : {pipeline} - {message}'


class AloPipelineSerializeError(AloError):
    """
    Raised when the object serialization(pickling) operation fails.
    Check if the object supports serialization by default when save(pickling)/load(unpickling).
    If the default serialization operation is not supported,
    implement the function to directly save/load the object in the path below context['model']['workspace'].
    """

    code = "ALO-PIP-010"
    fmt = '[{code}] Error : {message}'


class AloPipelineSummaryError(AloError):
    """
    Raised when create summary report.
    Check the values of result, score, note, probability
    """

    code = "ALO-PIP-011"
    fmt = '[{code}] Error : {message}'


class AloPipelineCompressError(AloError):
    """
    Raised when an error occurs while compressing/decompressing a file.
    Check the file extension to make sure it is a valid format.
    """

    code = "ALO-PIP-012"
    fmt = '[{code}] Compress/Decompress Error : {message}'


class AloPipelineArgumentError(AloError):
    """
    Raised when an argument value is invalid.
    Check the argument's settings.
    """

    code = "ALO-PIP-013"
    fmt = '[{code}] {message}'


#################
# Package Error
class AloPackageRequirementsError(AloError):
    """
    Raised when the 3rd party library cannot be installed.
    Check the library version or network.
    """

    code = "ALO-PAC-000"
    fmt = '[{code}] Found error when installing the package : {message}'


#################
# Sagemaker Error
class AloSagemakerInitError(AloError):
    """
    Raised when initialize various SageMaker-related config information as class variables.
    """

    code = "ALO-SAG-000"
    fmt = '[{code}] Message : {message}'


class AloSagemakerSetupError(AloError):
    """
    Raised when copy the elements required for docker build into the sagemaker directory for the given list of pipelines.
    """

    code = "ALO-SAG-001"
    fmt = '[{code}] Message : {message}'


class AloSagemakerBuildError(AloError):
    """
    Raised when docker build, ecr push, create s3 bucket for sagemaker.
    """

    code = "ALO-SAG-002"
    fmt = '[{code}] Message : {message}'


class AloSagemakerEstimatorError(AloError):
    """
    Raised when fit sagemaker estimator (execute on cloud resource).
    """

    code = "ALO-SAG-003"
    fmt = '[{code}] Message : {message}'


class AloSagemakerTrainError(AloError):
    """
    Raised when failed to download sagemaker trained model.
    """

    code = "ALO-SAG-004"
    fmt = '[{code}] Message : {message}'


#################
# Asset Error
class AloAssetSetupError(AloError):
    """
    Raised when failed to install asset.
    Check for duplicate step names and raise an error if any exist.
    """

    code = "ALO-ASS-000"
    fmt = '[{code}] Message : {message}'


class AloAssetRunError(AloError):
    """
    Raised when failed to user asset run.
    Check source code of user asset.
    """

    code = "ALO-ASS-001"
    fmt = '[{code}] Message : {message}'


class AloArtifactFileNoneError(AloError):
    """
    에러 코드
        {code}

    에러 정보
        Stage: {stage}

    에러 원인(제약 사항)
        추론 실행 결과 파일을 찾을 수 없습니다.
        저장할 수 있는 파일 개수는 2개 이하로 제한 되며, 반드시 1개 이상의 파일을 생성해야 합니다.

    조치 가이드
        참고 사항과 같이 한 개 이상의 파일이 저장될 수 있도록 기능을 구현하세요.

    참고 사항
        pipeline['artifact']['workspace'] 의 경로 정보를 참조하여 파일 생성
            >>> with open(os.path.join(pipeline['artifact']['workspace'], "inference1.csv"), "w") as f:
            >>>     f.write("inference1")
    """

    code = "ALO-ART-001"


class AloArtifactFileLimitError(AloError):
    """
    에러 코드
        {code}

    에러 정보
        - Stage:     {stage}
        - Artifacts: {files}

    에러 원인(제약 사항)
        추론 결과 파일은 2개를 초과하여 저장할 수 없습니다.
        저장할 수 있는 파일 개수는 2개 이하로 제한 되며, 반드시 1개 이상의 파일을 생성해야 합니다.

    조치 가이드
        에러 정보의 Artifacts 항목을 참고하여 파일 개수를 2개 이하로 제한하세요.
        불필요한 파일이 존재한다면 삭제하세요.

    참고 사항
        pipeline['artifact']['workspace'] 는 추론 파일을 저장하기 위한 경로 정보를 제공합니다.
            >>> print(pipeline['artifact']['workspace']) # /var/alo/workspace/train/output/result.csv

        pipeline['artifact']['workspace'] 값을 참조하여 파일 저장 로직을 구현한 부분이 있다면
        파일 저장 개수 제약 사항에 위반되지 않도록 불필요한 파일 저장 로직은 삭제하세요.
    """

    code = "ALO-ART-002"


class AloArtifactFileExtensionError(AloError):
    """
    에러 코드
        {code}

    에러 정보
        - Stage:    {stage}
        - Artifact: {files}

    에러 원인(제약 사항)
        허용되지 않는 파일 유형이 추론 결과 파일에 포함되어 있습니다.
        csv, jpg, jpeg, png, svg 파일 유형만 저장 가능합니다.

    조치 가이드
        에러 정보의 Artifacts 항목을 참고하여 허용되지 않는 파일 유형을 확인 후
        해당 파일 유형을 변경 또는 삭제하세요.

    참고 사항
        pipeline['artifact']['workspace'] 는 추론 파일을 저장하기 위한 경로 정보를 제공합니다.
            >>> print(pipeline['artifact']['workspace']) # /var/alo/workspace/train/output/result.csv
    """

    code = "ALO-ART-003"


class AloUserPipelineRuntimeError(AloError):
    """
    에러 코드
        {code}

    에러 정보
        - File:     {file}
        - Function: {function}
        - Message:  {message}

    에러 원인
        사용자 파이프라인 함수(python 코드) 실행중에 runtime 에러가 발생하였습니다.

    조치 가이드
        에러 정보의 File(python 코드)에서의 에러가 발생하였습니다.
        에러 메시지 상단의 Traceback을 참고하여 에러 발생 위치 및 exception 메시지를 확인 후
        코드를 수정하세요.

    참고 사항
        ALO에서 제공되는 context, pipeline 두 참조 dict 객체는 아래와 같은 정보를 가지고 있습니다.
        context: ALO 수행과 관련된 정보를 조회하거나, 저장할 수 있는 객체
            context['stage']: 현재 수행중인 단계(train 또는 inference) 리턴
                >>> print(context['stage']) # train or inference
            context['model']['workspace']: 모델 저장 경로 반환.
                >>> print(os.path.join(context['model']['workspace'], 'model.pickle')) # /var/alo/workspace/model/model.pickle
            context['model'][파일명]: 파일명(pickle 확장자 제외)에 해당하는 model을 메모리로 로딩 후 객체로 반환하거나, 저장(pickling 지원 대상만 해당)
                >>> context['model']['titanic'] = RandomForestClassifier(n_estimators=n_estimators, max_depth=5, random_state=1)  # titanic 이라는 이름으로 모델 객체를 파일로 저장
                >>> model = context['model']['titanic'] # titanic 이름으로 저장된 모델을 객체로 로딩 후 반환
        pipeline: train, inference 단계 수행시 dataset, artiface 파일 정보를 조회할 수 있는 객체
            pipeline['dataset']['workspace' 또는 파일명]: dataset 파일이 저장된 경로를 반환
                >>> print(os.path.join(pipeline['dataset']['workspace'], 'dataset.csv')) # /var/alo/workspace/train/dataset/file.csv
                >>> print(pipeline['dataset']['file.csv'])                               # /var/alo/workspace/train/dataset/file.csv (위와 동일)
            pipeline['artifact']['workspace']: artifact 파일을 저장하기 위한 경로를 반환
                >>> print(os.path.join(pipeline['artifact']['workspace'])) # /var/alo/workspace/train/output
    """

    code = "ALO-USR-001"


class AloUserPythonHandlerError(AloError):
    """
    에러 코드
        {code}

    에러 정보
        - File:     {file}
        - Function: {function}
        - Message:  {message}

    에러 원인
        사용자 파이프라인 함수(python 코드) 핸들러 정의시
        Message 내용과 같이 제약사항을 준수하지 않았습니다.

    조치 가이드
        아래 참고 사항을 참조하여 함수 정의 부분을 수정하세요.

    참고 사항
        ALO는 사용자 정의 함수를 실행 시점에 3가지 유형의 (객체)를 전달할 수 있습니다.

        1. context: ALO 수행과 관련된 정보를 조회하거나, 저장할 수 있는 객체
        2. pipeline: train, inference 단계 수행시 dataset, artiface 파일 정보를 조회할 수 있는 객체
        3. 사용자 정의 keyword arguments

        positional arguments 제약 사항
            - 인자가 없는 경우 : 아무런 정보도 넘겨줄 수 없음으로 dataset/model 정보를 가져올 수 없음
                >>> train():
                >>>     ...
            - 인자가 1개인 경우 : pipeline 객체를 전달
                >>> train(pipeline):
                >>>     print(pipeline['dataset']['file.csv'])
            - 인자가 2개인 경우 : context, pipeline 객체를 전달
                >>> train(context, pipeline):
                >>>     print(context['startAt'])
                >>>     print(pipeline['dataset']['file.csv'])
            - 인자가 3개 이상인 경우 :
                -> 오류(ALO-USR-002) 발생

        keyword arguments 제약 사항
            - keyword arguments 정의 하지 않는 경우
                >>> train(pipeline):
                >>>     ...
            - keyword arguments 개별 기본 값 정의
                >>> train(pipeline, x_columns=[], y_column=None, n_estimators=100):
                >>>     ...
            - keyword arguments dict형 정의
                >>> train(pipeline, **kwargs):
                >>>     kwargs['x_columns']
            - keyword arguments 와 solution.function.[함수명].argument 가 불일치 하는 경우
                -> 오류(ALO-USR-002) 발생

    """

    code = "ALO-USR-002"


class AloDatasetKeyNotFoundError(AloError):
    """
    에러 코드
        {code}

    에러 정보
        - Stage: {stage}
        - Key:   {key}

    에러 원인
        찾고자 하는 dataset 파일명({key})이 잘못되었거나,
        또는 파일이 누락된 경우에 해당합니다.

    조치 가이드
        | Key는 experimental_plan.yaml 파일의 solution.{stage}.dataset_uri에 설정된 경로 이하 또는 압축 파일 내에 폴더명 및 확장자명을 포함한 경로 형태여야 합니다.
        | Ex) pipeline['dataset']['path/train.csv']
        | 아래 각 항목을 점검하세요.

        1. experimental_plan.yaml의 solution.{stage}.dataset_uri 설정 유무 확인
        2. 압축 파일인 경우 파일 내부에 파일 포함 여부 확인
        3. 압축 파일 내부 폴더 이하에 파일이 존재하는 경우 경우
            >>> print(pipeline['dataset']['파일명.csv'])       # 오류 발생
            >>> print(pipeline['dataset']['폴더명/파일명.csv']) # /var/alo/workspace/train/dataset/폴더명/파일명.csv
        4. key에 파일 확장자명 누락된 경우
            >>> print(pipeline['dataset']['파일명'])           # 오류 발생
            >>> print(pipeline['dataset']['파일명.csv'])       # /var/alo/workspace/train/dataset/파일명.csv

    참고 사항
        pipeline['dataset'] 객체는 train/inference의 필요한 파일들에 대한 경로 정보를 제공합니다.
            >>> print(pipeline['dataset']['workspace']) # /var/alo/workspace/train/dataset
            >>> print(pipeline['dataset']['file.csv'])  # /var/alo/workspace/train/dataset/file.csv
    """

    code = "ALO-DTS-001"


class AloDatasetFileNotFoundError(AloError):
    """
    에러 코드
        {code}

    에러 정보
        - Stage: {stage}
        - Key:   {key}
        - File:  {file}

    에러 원인
        에러 정보의 Key 경로에 해당하는 File이 삭제되어 찾을 수 없습니다.

    조치 가이드
        사용자 파이프라인 함수(python 코드) 등록된 파일을 삭제한 로직이 없는지 검토가 필요합니다.

    참고 사항
        pipeline['dataset'] 객체는 train/inference의 필요한 파일들에 대한 경로 정보를 제공합니다.
            >>> print(pipeline['dataset']['workspace']) # /var/alo/workspace/train/dataset
            >>> print(pipeline['dataset']['file.csv'])  # /var/alo/workspace/train/dataset/file.csv
    """

    code = "ALO-DTS-002"


class AloModelFileNotFoundError(AloError):
    """
    에러 코드
        {code}

    에러 정보
        - Key:   {key}
        - File:  {file}

    에러 원인
        에러 정보의 Key에 해당하는 모델 파일(pickle)을 찾을 수 없습니다.

    조치 가이드
        key 및 모델 파일이 존재하는지 확인이 필요합니다.
        1. experimental_plan.yaml 파일의 solution.[stage].model_uri에 모델 파일 여부 확인
        2. 모델 파일의 확장자명이 .pkl 인지 여부 확인
        3. key명 오탈자 여부 확인
        4. key명에 확장자 포함 여부 확인

    참고 사항
        context['model'] 객체는 모델 파일(pickle)에 대한 정보를 제공하고 있습니다.
            >>> model = context['model']['titanic']     # titanic 이름으로 저장된 모델을 객체로 로딩 후 반환
            >>> model = context['model']['titanic.pkl'] # 오류 발생
    """

    code = "ALO-MDL-001"


class AloModelUnpicklingError(AloError):
    """
    에러 코드
        {code}

    에러 정보
        - Key:   {key}
        - File:  {file}

    에러 원인
        에러 정보의 Key에 해당하는 모델 파일(pickle)을 메모리를 로드(unpickling)할 수 없습니다.
        일부 모델 객체는 python pickling/unpickling 기능이 지원되지 않습니다.

    조치 가이드
        1. requirements.txt의 모델에 해당하는 library가 동일 버전인지 확인
            -> train시 pickle 파일로 저장시 library 버전과 unpickle시 library 버전이 상이한 경우
            Ex) scikit-learn==1.4.0 의 모델 객체를 pickle 파일로 저장 후 scikit-learn==1.5.0 버전으로 unpickle 시 오류 발생
        2. 모델 객체에 대해 pickling/unpickling을 지원하지 않는 경우
            A. pickling/unpickling 기능을 직접 구현하는 방법
                >>> import os
                >>> def pickle(context, model: object):
                >>>     # model_bytes =  ...(변환 로직 구현)
                >>>     with open(os.path.join(context['model']['workspace'], "my_custom_model.pkl"), "wb") as f:  # 모델 경로 이하에 저장
                >>>         f.write(model_bytes)
                >>> model = None
                >>> def unpickle(context):
                >>>     global model  # 모델 객체 생성에 대한 cost를 줄이기 위해 global 변수를 통해 재사용
                >>>     # model_bytes =  ...(변환 로직 구현)
                >>>     if model is None:
                >>>         with open(os.path.join(context['model']['workspace'], "my_custom_model.pkl"), "rb") as f:  # 모델 경로 이하에 저장
                >>>             model_bytes = f.read()
                >>>             model = ...(변환 로직 구현)
                >>>     return model
            B. 모델 객체의 parameter 만 pickle 파일로 저장 후 모델 객체의 parameter로 전달하는 방법
                >>> def train(context: dict, pipeline: dict):
                >>>     context['model']['titanic_param'] = ("n_estimators": 100, "max_depth": 5, "random_state": 1)  # model parameter pickle로 저장
                >>>     model = RandomForestClassifier(**context['model']['titanic_param'])
                >>> model = None
                >>> def inference(context: dict, pipeline: dict):
                >>>     global model  # 모델 객체 생성에 대한 cost를 줄이기 위해 global 변수를 통해 재사용
                >>>     if model is None:
                >>>         titanic_param = context['model']['titanic_param']
                >>>         model = RandomForestClassifier(**titanic_param)

    참고 사항
        ALO에서는 python pickle library 를 사용하여
        pickling(객체 -> 파일) 또는 unpickling(파일 -> 객체) 하는 기능을 기본 지원하고 있습니다.
            >>> def train(context: dict, pipeline: dict):
            >>>     model = RandomForestClassifier(n_estimators=100, max_depth=5, random_state=1)
            >>>     context['model']['titanic'] = model  # 모델을 파일로 pickling
            >>>
            >>> def inference(context: dict, pipeline: dict):
            >>>     model = context['model']['titanic']  # 파일을 RandomForestClassifier 객체로 unpickling

        일부 모델 객체들에 대해서는 python pickle 기능이 적용되지 않음으로
        조치 가이드 2번 항목을 참고하세요.

    """

    code = "ALO-MDL-002"


class AloModelPicklingError(AloError):
    """
    에러 코드
        {code}

    에러 정보
        - Key:   {key}
        - File:  {file}

    에러 원인
        에러 정보의 Key에 해당하는 객체(모델)를 파일(pickle)로 저장(pickling)할 수 없습니다.
        일부 모델 객체는 python pickling/unpickling 기능이 지원되지 않습니다.

    조치 가이드
        1. 사용중인 library에서의 pickling 제공 여부를 확인하거나, 제 3의 library(Ex. fickling 등) 적용
        2. 모델 객체에 대해 pickling/unpickling을 지원하지 않는 경우
            A. pickling/unpickling 기능을 직접 구현하는 방법
                >>> import os
                >>> def pickle(context, model: object):
                >>>     # model_bytes =  ...(변환 로직 구현)
                >>>     with open(os.path.join(context['model']['workspace'], "my_custom_model.pkl"), "wb") as f:  # 모델 경로 이하에 저장
                >>>         f.write(model_bytes)
                >>> model = None
                >>> def unpickle(context):
                >>>     global model  # 모델 객체 생성에 대한 cost를 줄이기 위해 global 변수를 통해 재사용
                >>>     # model_bytes =  ...(변환 로직 구현)
                >>>     if model is None:
                >>>         with open(os.path.join(context['model']['workspace'], "my_custom_model.pkl"), "rb") as f:  # 모델 경로 이하에 저장
                >>>             model_bytes = f.read()
                >>>             model = ...(변환 로직 구현)
                >>>     return model
            B. 모델 객체의 parameter 만 pickle 파일로 저장 후 모델 객체의 parameter로 전달하는 방법
                >>> def train(context: dict, pipeline: dict):
                >>>     context['model']['titanic_param'] = ("n_estimators": 100, "max_depth": 5, "random_state": 1)  # model parameter pickle로 저장
                >>>     model = RandomForestClassifier(**context['model']['titanic_param'])
                >>> model = None
                >>> def inference(context: dict, pipeline: dict):
                >>>     global model  # 모델 객체 생성에 대한 cost를 줄이기 위해 global 변수를 통해 재사용
                >>>     if model is None:
                >>>         titanic_param = context['model']['titanic_param']
                >>>         model = RandomForestClassifier(**titanic_param)

    참고 사항
        ALO에서는 python pickle library 를 사용하여
        pickling(객체 -> 파일) 또는 unpickling(파일 -> 객체) 하는 기능을 기본 지원하고 있습니다.
            >>> def train(context: dict, pipeline: dict):
            >>>     model = RandomForestClassifier(n_estimators=100, max_depth=5, random_state=1)
            >>>     context['model']['titanic'] = model  # 모델을 파일로 pickling
            >>>
            >>> def inference(context: dict, pipeline: dict):
            >>>     model = context['model']['titanic']  # 파일을 RandomForestClassifier 객체로 unpickling

        일부 모델 객체들에 대해서는 python pickle 기능이 적용되지 않음으로
        조치 가이드 내용을 참고하세요.
    """

    code = "ALO-MDL-003"


class AloModelTrainFileNotFoundError(AloError):
    """
    에러 코드
        {code}

    에러 정보
        Phase: {phase}

    에러 원인(제약 사항)
        학습(Train) pipline 에서 모델 관련 output 파일이 없습니다.
        조치 가이드와 같이 1개 이상의 모델 관련 파일을 저장하세요.

    조치 가이드
        1. 모델 객체를 pickle 파일로 저장
            >>> model = RandomForestClassifier(n_estimators=n_estimators, max_depth=5, random_state=1)
            >>> context['model']['titanic'] = model

        2. config 정보(dict)를 파일로 저장
            >>> context['model']['model_config'] = dict(n_estimators=100, max_depth=5, random_state=1)

    참고 사항
        context['model'] 객체는 모델 관련 정보를 제공하고 있습니다.
        예제) 모델 관련 객체 또는 설정 정보를 메모리로 로드
            >>> model = context['model']['titanic']     # titanic 이름으로 저장된 모델을 객체로 로딩 후 반환

        예제) 모델 관련 객체 또는 설정 정보를 파일로 쓰기
            >>> context['model']['titanic'] = model     # model 객체를 titanic 파일 정보로 저장

        예제) 사용자 정의 model 관련 파일 생성시 context['model']['workspace'] 경로 정보를 참고하여 파일 생성
            >>> with open(os.path.join(context['model']['workspace'], "titanic.pkl"), "wb") as f:  # 모델 경로 이하에 저장 직접 파일로 저장
            >>>     f.write(model_bytes)  # model 관련 정보를 파일로 저장

    """

    code = "ALO-MDL-004"


class AloModelFileStorageLimitError(AloError):
    """
    에러 코드
        {code}

    에러 정보
        - Property : {property}
        - Message  : {message}

    에러 원인(제약 사항)
        solution.[train|inference] 의 속성 dataset_uri, model_uri 에 로컬 경로 설정시
        제약 사항을 위반하였습니다.

    조치 가이드
        uri 속성에 파일의 위치 정보(S3, 로컬 경로) 작성시
        아래 참고 사항과 같은 제약사항이 존재합니다.

    참고 사항
        - experimental_plan.yaml 상에 solution.train.dataset_uri은 및 solution.inference.dataset_uri의 경로는 폴더까지만 입력합니다.
          string, list type 입력이 가능합니다.
            >>> dataset_uri: sample_dataset/                           # 정상
            >>> dataset_uri: ['sample_dataset/']                       # 정상
            >>> dataset_uri: ['sample_dataset1/', 'sample_dataset2/']  # 정상
            >>> dataset_uri: sample_dataset/dataset.csv                # 오류 - 폴더가 아닌 파일 경로 입력 불가

        - 로컬 경로의 파일 지정시 experimental_plan.yaml 파일 기준 상대 경로로 작성해야 합니다.
            >>> Ex) 폴더 트리 인 경우
            >>> ./
            >>> ├── experimental_plan.yaml
            >>> ├── sample_dataset/
            >>> │   ├── train_dataset.csv
            >>> │   └── train_dataset1.csv
            >>>     ...
            >>> └── titanic.py
            >>> dataset_uri: sample_dataset/                            # 정상
            >>> dataset_uri: /home/user/alo/titanic/sample_dataset/     # 오류 - 절대 경로로 지정함
    """

    code = "ALO-MDL-005"


class AloContextWrongKeyTypeError(AloError):
    """
    에러 코드
        {code}

    에러 정보
        - Key: {key}
        - Type: {type}

    에러 원인(제약 사항)
        사용자 정의 key로 등록 가능한 데이터 유형은 문자열만 가능합니다.
        '{type}' 데이터 유형은 key로 사용할 수 없습니다.

    조치 가이드
        >>> context['model'][1] = 'abc'    # 오류 발생. 숫자 1은 key로 사용 불가
        >>> context['model']['1'] = 'abc'  # 정상 수행
        상기 라인을 수정 또는 삭제하세요.

    참고 사항
        시스템에서 제공되는 데이터 key는 모두 문자형 타입입니다.
            - context['model']['workspace']
            - context['model']['example_model']
            - pipeline['dataset']['workspace']
            - pipeline['dataset']['train.csv']
            - pipeline['artifact']['workspace']
    """

    code = "ALO-CTX-001"


class AloContextNotAllowKeyError(AloError):
    """
    에러 코드
        {code}

    에러 정보
        Key: {key}

    에러 원인(제약 사항)
        '{key}' key는 시스템에서 제공되는 key명이며, 읽기만 가능한 key 입니다.
        key에 해당하는 값을 업데이트 할 수 없습니다.

    조치 가이드
        >>> context[...]['{key}'] = '값'
        >>> 또는
        >>> pipeline[...]['{key}'] = '값'
        상기 라인을 수정 또는 삭제하세요.

    참고 사항
        시스템에서 제공되는 일부 key는 임의로 수정할 수 없습니다.
        값 변경 불가한 key 목록입니다.
            - context['model']['workspace']
            - pipeline['dataset']['workspace']
            - pipeline['artifact']['workspace']
    """

    code = "ALO-CTX-002"


class AloContextInvalidStrError(AloError):
    """
    에러 코드
        {code}

    에러 정보
        Invalid Key: {invalidKey}

    에러 원인(제약 사항)
        '{invalidKey}'에는 사용 불가능한 문자가 포함되어 있습니다.
        조치 가이드 및 참고 사항을 참고하여 문자 또는 문자열을 변경하세요.

    조치 가이드
        소,중,대 괄호 및 특수문자는 사용 불가능하며,
        아래와 같이 사용가능한 문자열로 변경하세요.
            >>> context['model']['key\\test.csv'] = 'test'  # 역슬래쉬(\) 사용에 의한 오류
            >>> context['model']['key/test.csv'] = 'test'  # 정상

    참고 사항
        허용 가능 문자 : 한글, 영문, 숫자, _, -, /, (공백)
        허용 문자열 정규 표현식 : {allowedRegex}
    """

    code = "ALO-CTX-003"


class AloSummaryProbabilityTypeError(AloError):
    """
    에러 코드
        {code}

    에러 정보
        - Key  : summary['probability']
        - Type : {keyType}

    에러 원인(제약 사항)
        파이프라인 함수 실행 결과에 대한 summary의 probability는 dict만 허용됩니다.
        조치 가이드 또는 참고 사항과 같이 수정하세요.

    조치 가이드
        probability 값을 dict 유형으로 변경하세요.
            >>> summary['probability'] = "dead: 0.2, survived: 0.8"      # 오류(value의 data type이 dit 가 아님)
            >>> summary['probability'] = {{"dead": 0.2, "survived": 0.8}}  # 정상

    참고 사항
        파이프라인 함수의 summary는
        아래 예제와 같이 return을 통해 결과를 전달할 수 있습니다.
            >>> def inference(context: dict, pipeline: dict, x_columns=[]):
            >>> ...
            >>>     return {{
            >>>         'summary': {{
            >>>             'result': f"#survived:100 / #total:20",
            >>>             'score': 0.1,
            >>>             'note': "Score means titanic survival ratio",
            >>>             'probability': {{"dead": 0.2, "survived": 0.8}}
            >>>         }}
            >>>     }}
    """

    code = "ALO-SMM-001"


class AloSummaryProbabilityDataTypeError(AloError):
    """
    에러 코드
        {code}

    에러 정보
        - Key        : summary['probability'][{key}]
        - Key Type   : {keyType}
        - Value      : {value}
        - Value Type : {valueType}

    에러 원인(제약 사항)
        파이프라인 함수 실행 결과에 대한 summary의 probability의 item은 key(문자):value(정수 또는 실수) 형태만 허용됩니다.
        조치 가이드 또는 참고 사항과 같이 수정하세요.

    조치 가이드
        아래 예제와 같이
        summary['probability']['key'] = 정수 또는 실수
        key:value 에 대한 데이터 유형을 확인 후 수정하세요.
            >>> summary['probability'][100] = 10        # 오류(key가 문자형이 아님)
            >>> summary['probability']['dead'] = '0.2'  # 오류(value가 숫자(정수 또는 실수)가 아님)
            >>> summary['probability']['dead'] = 0.2    # 정상

    참고 사항
        파이프라인 함수의 summary는
        아래 예제와 같이 return을 통해 결과를 전달할 수 있습니다.
            >>> def inference(context: dict, pipeline: dict, x_columns=[]):
            >>> ...
            >>>     return {{
            >>>         'summary': {{
            >>>             'result': f"#survived:100 / #total:20",
            >>>             'score': 0.1,
            >>>             'note': "Score means titanic survival ratio",
            >>>             'probability': {{"dead": 0.2, "survived": 0.8}}
            >>>         }}
            >>>     }}
    """

    code = "ALO-SMM-002"


class AloSummaryProbabilityValueSumError(AloError):
    """
    에러 코드
        {code}

    에러 정보
        - Values        : {values}
        - Sum of Values : {sumValues}

    에러 원인(제약 사항)
        파이프라인 함수 실행 결과에 대한 summary의 probability value의 합은 1.0 과 같아야만 합니다.
        조치 가이드 또는 참고 사항과 같이 수정하세요.

    조치 가이드
        summary['probability'].values() 의 합이 1이 되도록 값을 수정하세요.
            >>> summary['probability'] = {{"dead": 0.1, "survived": 0.8}}  # 오류(합계가 1.0과 같지 않음)
            >>> summary['probability'] = {{"dead": 0.2, "survived": 0.8}}  # 정상

    참고 사항
        파이프라인 함수의 summary는
        아래 예제와 같이 return을 통해 결과를 전달할 수 있습니다.
            >>> def inference(context: dict, pipeline: dict, x_columns=[]):
            >>> ...
            >>>     return {{
            >>>         'summary': {{
            >>>             'result': f"#survived:100 / #total:20",
            >>>             'score': 0.1,
            >>>             'note': "Score means titanic survival ratio",
            >>>             'probability': {{"dead": 0.2, "survived": 0.8}}
            >>>         }}
            >>>     }}
    """

    code = "ALO-SMM-003"


class AloExperimentalPlanFileNotFoundError(AloError):
    """
    에러 코드
        {code}

    에러 정보
        File Not Found : {fileNotFound}

    에러 원인(제약 사항)
        experimental_plan.yaml 파일을 찾을 수 없습니다.
        조치 가이드 내용과 같이 파일 경로 확인 또는 옵션을 지정하세요.

    조치 가이드
        experimental_plan.yaml 파일 우선순위는 아래와 같습니다.

        1. --config 옵션 지정시
          Ex) alo --config /mellerikat/experimental_plan.yaml
        2. git repository url 옵션 지정시
          Ex) alo --git.url http://mellerikat.com/repository/titanic.git --git.branch main
          (titanic 이하 폴더에 experimental_plan.yaml 파일이 존재해야만 함)
        3. 사용자 현재 경로 이하에 experimental_plan.yaml 파일이 존재하는 경우
        4. python alo package에 설치된 경로에 solution/experimental_plan.yaml 파일이 존재하는 경우

    참고 사항
        사용자 python 파일을 로딩하기 위해서 experimental_plan.yaml 파일이 존재하는 폴더 이하에 py파일 생성 후
        yaml의 solution.function 항목에 실행 대상 함수를 "모듈명.함수명" 형식으로 정의하세요.
            >>> Ex) 폴더 트리
            >>> ./
            >>> ├── experimental_plan.yaml
            >>> └── titanic.py
        Ex) experimental_plan.yaml 사용자 함수 정의 작성 방법
            >>> function:
            >>>   preprocess:                  # 함수 이름
            >>>     def: titanic.preprocess    # 모듈명.함수명(python 파일명.def 함수명)
            >>>   train:
            >>>     def: titanic.train
            >>>     argument:
            >>>       x_columns: [ 'Pclass', 'Sex', 'SibSp']
            >>>       y_column: Survived
            >>>   inference:
            >>>     def: titanic.inference
    """

    code = "ALO-EPP-001"


class AloSolutionStreamApiStatusCodeError(AloError):
    """
     에러 코드
         {code}

     에러 정보
         - Type        : {type}
         - API         : {api}
         - Status Code : {statusCode}
         - Message     : {message}

     에러 원인
        API 호출 결과 정상적인 응답을 받지 못하였습니다.
        메시지 정보를 읽고 관련 내용을 직접 조치하거나,
        조치 불가능한 경우 AI Conductor 관리자에게 문의하세요.
        [주의] Solution 등록 도중에 AI Conductor 웹 브라우저에 로그인하면 등록이 중단됩니다.
     """

    code = "ALO-SSA-001"


class AloSolutionStreamFileNotFoundError(AloError):
    """
    에러 코드
        {code}

    에러 정보
        File : {file}

    에러 원인
        File 정보를 찾을 수 없습니다.
        아래 참고 사항을 읽고 누락된 설정 파일이 없는지 확인하세요.
        ALO 관리자에게 문의하세요.

    참고 사항
        ALO 실행 및 솔루션 등록을 위해서는 아래 예제와 같이 파일 구성이 필요합니다.
            >>> ./
            >>> ├── experimental_plan.yaml      # ALO 실행 환경 설정 파일
            >>> ├── inference/
            >>> │   └── dataset.csv             # titanic.의 추론용 테스트 파일
            >>> ├── titanic.py                  # ALO 실행 소스 코드
            >>> ├── train/
            >>> │   └── dataset.csv             # titanic.의 학습용 테스트 파일
            >>> └── setting/                    # 솔수션 등록 관련 환경 설정
            >>>     ├── infra_config.yaml           # mellerikat 인프라 구성 정보
            >>>     └── solution_info.yaml          # 솔루션 등록 환경 설정 정보
     """

    code = "ALO-SSA-002"


class AloSolutionStreamAwsCredentialsError(AloError):
    """
     에러 코드
         {code}

     에러 정보
         - AWS Profile : {awsProfile}
         - Service Name: {serviceName}

     에러 원인
        AWS Profile 을 찾을 수 없거나, 또는 권한이 없을 수 있습니다.

    조치 가이드
        AWS Profile 이 정상적으로 설정되 있는지 아래 명령어를 통해서 확인하세요.
            >>> $ aws configure list --profile your_profile
            >>>           Name                    Value             Type    Location
            >>>           ----                    -----             ----    --------
            >>>        profile             your_profile             None    None
            >>>     access_key     ******************** shared-credentials-file
            >>>     secret_key     ******************** shared-credentials-file
            >>>         region           ap-northeast-2      config-file    ~/.aws/config

        AWS Profile 을 확인할 수 없다면, 아래 명령어를 통해서 설정하세요.
            >>> $ aws configure
            >>> AWS Access Key ID [None]: AKIAIOSFODNN7EXAMPLE
            >>> AWS Secret Access Key [None]: wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY
            >>> Default region name [None]: us-west-2
            >>> Default output format [None]: json

        권한이 누락된 경우 AWS Admin에게 문의하세요.
     """

    code = "ALO-SSA-003"


class AloSolutionStreamApiCallError(AloError):
    """
     에러 코드
         {code}

     에러 정보
         - API   : {api}
         - Error : {error}

     에러 원인
        API를 호출할 수 없습니다.
        에러 정보와 함께 AI Conductor 관리자에게 문의하세요.
        [주의] Solution 등록 도중에 AI Conductor 웹 브라우저에 로그인하면 등록이 중단됩니다.
     """

    code = "ALO-SSA-004"


class AloSolutionStreamLoginFailError(AloError):
    """
     에러 코드
         {code}

     에러 정보
         Message : {message}

     에러 원인
        사용자 정보가 불일치 하거나, 부족합니다.
        계정 정보를 확인하세요.
     """

    code = "ALO-SSA-005"


class AloSolutionStreamWrongNameError(AloError):
    """
     에러 코드
         {code}

     에러 정보
         - Name   : {name}
         - Message: {message}

     에러 원인(제약 사항)
        Solution 생성시 아래 제약사항을 위반하였습니다.
        명명 규칙 및 신규/수정 설정값을 참고 하세요.

    조치 가이드
        아래 참고 사항을 통해서 이름 작성 규칙 및 신규 생성 여부 설정값을 확인하세요.

    참고 사항
        신규 생성시 Solution name 작성 규칙
            - 동일 이름 생성 불가
            - 길이는 100 byte 를 이내
            - 영문 소문자, 대문자, 숫자, -(대시) 문자만 허용
            - 공백 불가

        setting/solution_info.yaml의 solution_update 속성
            - false: 신규 생성을 의미
            - true: 기존 항목을 수정을 의미
     """

    code = "ALO-SSA-006"


class AloSolutionStreamFileOpenFailError(AloError):
    """
     에러 코드
         {code}

     에러 정보
         - File : {file}
         - Message: {message}

     에러 원인
        File 열거나 저장할 수 없습니다.
        ALO 관리자에게 문의하세요.
     """

    code = "ALO-SSA-007"


class AloSolutionStreamStorageApiError(AloError):
    """
     에러 코드
         {code}

     에러 정보
         - Type    : {type}
         - Source  : {source}
         - Target  : {target}
         - Message : {message}

     에러 원인
        솔루션 등록/수정/삭제 작업을 위한 파일 관리 작업에 실패하였습니다.
        Source/Target 경로가 잘못되었가나, 권한이 없을 수 있습니다.

        setting/infra_config.yaml 파일의 AWS_KEY_PROFILE 설정 항목 확인 후
        해당 profile명이 업로드 권한을 가지고 있는 것인지 확인이 필요합니다.

    참고 사항
        setting/infra_config.yaml 의 AWS_KEY_PROFILE 명을 통해서 솔루션 등록 작업을 이루어지게 됩니다.

        AWS_KEY_PROFILE: aws credentials profile 명
            - 미설정시 : default OS 환경 변수 또는 부여된 role을 사용
            - 설정시 : aws credentials profile 명 사용
     """

    code = "ALO-SSA-008"


class AloSolutionStreamDockerFailError(AloError):
    """
     에러 코드
         {code}

     에러 정보
         - Phase   : {phase}
         - Message : {message}

     에러 원인
        Dockerfile 생성 작업 중 오류가 확인되었습니다.
        ALO 관리자에게 에러 메시지와 함께 문의하세요.

     참고사항
        solution_info.yaml에 gpu 사용을 true로 설정하신 경우, 사전에
        alo docker --gpu 를 통해 gpu 용 dockerfile 작성을 수행해야 합니다.

     """

    code = "ALO-SSA-009"


class AloSolutionStreamEcrApiError(AloError):
    """
     에러 코드
         {code}

     에러 정보
         - Type       : {type}
         - Repository : {repository}
         - Message    : {message}

     에러 원인
        솔루션 이미지 관리를 위한 Container Registry 작업에 실패하였습니다.
        권한이 없거나, 잘못된 접근일 수도 있습니다.
        에러 메시지를 확인하세요.

        setting/infra_config.yaml 파일의 AWS_KEY_PROFILE 설정 항목 확인 후
        해당 profile명이 업로드 권한을 가지고 있는 것인지 확인이 필요합니다.

        오류 원인 및 조치가 어려운 경우 ALO 담당자에게 문의하세요.

    참고 사항
        setting/infra_config.yaml 의 AWS_KEY_PROFILE 명을 통해서 솔루션 등록 작업을 이루어지게 됩니다.

        AWS_KEY_PROFILE: aws credentials profile 명
            - 미설정시 : default OS 환경 변수 또는 부여된 role을 사용
            - 설정시 : aws credentials profile 명 사용
     """

    code = "ALO-SSA-010"


class AloSolutionStreamCodeBuildError(AloError):
    """
     에러 코드
         {code}

     에러 정보
         Message : {message}

     에러 원인
        Code Build 관련 작업 수행 중 오류가 발생했습니다.
        ALO 관리자에게 에러 메시지와 함께 문의하세요.
     """

    code = "ALO-SSA-011"


class AloSolutionStreamRegisterLimitError(AloError):
    """
     에러 코드
         {code}

     에러 정보
         Message: {message}

     에러 원인(제약 사항)
        솔루션 등록 시 아래 참고 사항과 같은 제약 사항이 존재합니다.
        에러 정보의 message를 확인 후 조치하세요.

    조치 가이드
        아래 참고 사항의 제약조건이 위배되지 않도록 수정하세요.

    참고 사항
        - alo 를 통해 1회 이상 실행 후 등록할 수 있습니다.
            -> `alo` 또는 `alo run` 명령어 실행이 필요
        - 이하 사용자 경로 이하에 `alo` 폴더는 포함될 수 없습니다.
            -> 폴더명을 변경 또는 삭제
     """

    code = "ALO-SSA-012"


class AloSolutionStreamGpuLibLimitError(AloError):
    """
     에러 코드
         {code}

     에러 정보
         Message: {message}

     에러 원인(제약 사항)
        GPU 사용을 위해서는 아래와 같은 제약사항이 존재합니다.
        에러 정보의 message를 확인 후 참고 사항과 같이 내용을 수정하세요.

    조치 가이드
        아래 참고 사항의 제약조건이 위배되지 않도록 수정하세요.

    참고 사항
        - GPU를 사용하기 위해서는 tensorflow, torch 둘 중 하나의 버전을 사용해야만 합니다.
            -> requirements 에 tensorflow==2.15.0 과 같이 작성하세요.
        - tensorflow, torch의 상세 버전을 받드시 명시적으로 기입하세요.
              >>> tensorflow          # 오류 : 버전 정보 누락
              >>> tensorflow==2.15.0  # 정상
        - lib별 지원 버전
            - tensorflow : 2.12 ~ 2.15
            - torch      : 2.0, 2.1
            -> 상기 버전외 GPU 는 지원하지 않습니다.
            -> 버전에 대한 추가 지원이 필요한 경우 ALO 관리자에게 문의하세요.
     """

    code = "ALO-SSA-013"

class AloSolutionStreamDefaultSummaryError(AloError):
    """
    에러 코드
        {code}

    에러 정보
        File    : {file}
        Message : {message}

    에러 원인
        Summary 파일이 기본값을 가지고 있어 솔루션 등록이 불가능합니다.
        Train/Inference가 정상적으로 실행되지 않았거나, 실행 결과가 저장되지 않았을 수 있습니다.

    조치 가이드
        1. Train/Inference pipeline에서 inference summary를 제작하도록 하세요.
        2. Summary 파일의 각 필드가 아래 기본값과 같은지 확인하세요:
           - probability: {{}}
           - result: None
           - score: None
        3. 기본값과 같은 경우 문제가 생길 수 있습니다

    참고 사항
        Summary 파일은 Train/Inference 실행 시 사용자가 제작하지 않으면 자동으로 default 값으로 생성됩니다.
        support_labeling을 사용하시려면 default 값이 아니고 사용자가 직접 summary를 제작하셔야 합니다
    """

    code = "ALO-SSA-014"

class AloSolutionStreamDockerCleanupError(AloError):
    """
    에러 코드
        {code}

    에러 정보
        File    : {file}
        Message : {message}

    에러 원인
        Docker 리소스를 정리하는 동안 에러가 발생한 경우

    조치 가이드
        1. LOG 파일을 확인 후 ALO 관리자엑게 문의 하세요

    참고 사항
    """
    code = "ALO-SSA-015"

class AloSolutionStreamArtifactsCleanupError(AloError):
    """
    에러 코드
        {code}

    에러 정보
        File    : {file}
        Message : {message}

    에러 원인
        아티팩트를 정리하는 동안 에러가 발생한 경우

    조치 가이드
        1. LOG 파일을 확인 후 ALO 관리자엑게 문의 하세요

    참고 사항
    """
    code = "ALO-SSA-016"

class AloSolutionStreamS3DataCleanupError(AloError):
    """
    에러 코드
        {code}

    에러 정보
        File    : {file}
        Message : {message}

    에러 원인
        S3 데이터를 정리하는 동안 에러가 발생한 경우

    조치 가이드
        1. LOG 파일을 확인 후 ALO 관리자엑게 문의 하세요

    참고 사항
    """
    code = "ALO-SSA-017"


AloErrors = {code: cls for code, cls in AloError.codes.items()}


import os
import sys
import logging
import random
import pandas as pd
import numpy as np
from datetime import datetime, timezone
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import precision_score
from sklearn.model_selection import train_test_split

def preprocess(pipeline: dict):
    logger = pipeline['logger']
    logger.debug("preprocess")

def train(pipeline: dict, x_columns=[], y_column=None, n_estimators=100):
    logger = pipeline['logger']
    logger.debug("train")
    file_list = os.listdir(pipeline['dataset']['workspace'])
    csv_files = [file for file in file_list if file.endswith('.csv')]
    
    for csv_file in csv_files:
        file_path = os.path.join(pipeline['dataset']['workspace'], csv_file)
        df = pd.read_csv(file_path)
        logger.debug("\n%s", df)
        X = pd.get_dummies(df[x_columns])
        X_train, X_test, y_train, y_test = train_test_split(X, df[y_column], test_size=0.2, random_state=42)

        model = RandomForestClassifier(n_estimators=n_estimators, max_depth=5, random_state=1)
        model.fit(X_train, y_train)

        y_pred = model.predict(X_test)
        precision = precision_score(y_test, y_pred, average='macro')
        logger.debug("y_pred\n%s", y_pred)
        pipeline['model']['n100_depth5'] = model  # save model
    
    return {
        'summary': {
            'result': f'precision: {precision}',
            'note': f'Test Titanic-demo (date: {datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M:%S")})',
            'score': random.uniform(0.1, 1.0),
        }
    }


def inference(pipeline: dict, x_columns=[]):
    logger = pipeline['logger']
    logger.debug("inference")
    model = pipeline['model']['n100_depth5']
    
    file_list = os.listdir(pipeline['dataset']['workspace'])
    csv_files = [file for file in file_list if file.endswith('.csv')]

    for csv_file in csv_files:
        file_path = os.path.join(pipeline['dataset']['workspace'], csv_file)
        df = pd.read_csv(file_path)
        logger.debug("\n%s", df)
        X = pd.get_dummies(df[x_columns])

        # load trained model
        predict_class = model.predict(X)
        predict_proba = model.predict_proba(X)

        result = pd.concat([df, pd.DataFrame(predict_class, columns=['predicted'])], axis=1)
        print(result)

        # result csv 저장
        result.to_csv(f"{pipeline['artifact']['workspace']}/result.csv")
        logger.debug("Save : %s", f"{pipeline['artifact']['workspace']}/result.csv")

        # summary
        num_survived = len(result[result['predicted'] == 1])
        num_total = len(result)
        survival_ratio = num_survived / num_total
        avg_proba = np.average(predict_proba, axis=0)
        avg_proba_survived = avg_proba[1].item()  # float
        avg_proba_dead = avg_proba[0].item()

    return {
        'extraOutput': '',
        'summary': {
            'result': f"#survived:{num_survived} / #total:{num_total}",
            'score': round(survival_ratio, 3),
            'note': "Score means titanic survival ratio",
            'probability': {"dead": avg_proba_dead, "survived": avg_proba_survived}
        }
    }


# ----------------------------------------------------------------------
# Copyright (C) 2024, mellerikat. LGE
# ----------------------------------------------------------------------

"""
ALO
"""

import os
import json
import redis
import uuid
import shutil
import pickle
import tarfile
import re
import yaml
import pkg_resources  # todo deprecated. should be fixed.
import zipfile
import psutil
import glob
import hashlib
import inspect
from abc import ABCMeta, abstractmethod
from enum import Enum
from copy import deepcopy
from datetime import datetime
from collections import OrderedDict
from pathlib import Path
from functools import wraps
from threading import Thread
from pytz import timezone

from alo.model import settings
from alo.exceptions import AloError, AloErrors
from alo.logger import LOG_PROCESS_FILE_NAME, create_pipline_handler, log_start_finish
from alo.model import load_model, SolutionMetadata, update_storage_credential, EXP_FILE_NAME, copytree
from alo.utils import ResourceProfile, print_copyright, print_table
from alo.__version__ import __version__, COPYRIGHT


logger = settings.logger
TRAIN = 'train'
INFERENCE = 'inference'
MODES = [TRAIN, INFERENCE]
LOG_PIPELINE_FILE_NAME = "pipeline.log"
ARTIFACT = 'artifact'
HISTORY_FOLDER_FORMAT = "%Y%m%dT%H%M%S.%f"
HISTORY_PATTERN = re.compile(r'([0-9]{4}[0-9]{2}[0-9]{2}T[0-9]{2}[0-9]{2}[0-9]{2}.[0-9]{6})($|-error$)')
RUN_PIPELINE_NAME = '__pipeline_names__'
RESULT_INFO_FILE = 'result_info.json'


def extract_file(file_paths: list, destination: str):
    """ 지정된 경로에 압축 파일을 해제합니다.

    Args:
        file_paths: 파일 경로 목록
        destination: 압축 해제된 파일을 저장할 경로

    Raises:
        AloErrors: ALO-PIP-012

    """
    for file_path in file_paths:
        try:
            if file_path.lower().endswith(('.tar.gz', '.tgz')):
                with tarfile.open(file_path) as file:
                    file.extractall(os.sep.join(file_path.split(os.sep)[:-1]))
                    logger.debug("[FILE] Extract %s: %s ", file_path, file.getnames())
            elif file_path.lower().endswith('.zip'):
                with zipfile.ZipFile(file_path) as file:
                    file.extractall(os.sep.join(file_path.split(os.sep)[:-1]))
                    logger.debug("[FILE] Extract %s: %s ", file_path, file.namelist())
        except Exception as e:
            raise AloErrors['ALO-PIP-012'](file_path) from e


def tar_dir(_path, _save_path, last_dir):
    """ compress directory as tar.gz

    Args:
        _path       (str): path tobe compressed
        _save_path  (str): tar.gz file save path
        last_dir   (str): last directory for _path

    Returns: -

    """
    tar = tarfile.open(_save_path, 'w:gz')
    for root, dirs, files in os.walk(_path):
        base_dir = root.split(last_dir)[-1] + '/'
        for file_name in files:
            # Arcname: Compress starting not from the absolute path beginning with /home,
            # but from train_artifacts/ or models/
            tar.add(os.path.join(root, file_name), arcname=base_dir + file_name)
    tar.close()


def zip_dir(_path, _save_path):
    """ compress directory as zip

    Args:
        _path       (str): path tobe compressed
        _save_path  (str): zip file save path
        _last_dir   (str): last directory for _path

    Returns: -

    """
    # remove .zip extension
    _save_path = os.path.splitext(_save_path)[0]
    shutil.make_archive(_save_path, 'zip', _path)


def add_logger_handler(func):
    """ 데코레이터 함수

    특정 함수의 로그를 별로도 분리하기 위해 default logger에
    파일 핸들러를 추가 후 자동 제거 합니다.

    Args:
        func    (function): original function

    Returns:
        wrapper (function): wrapped function

    """
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        _handler = create_pipline_handler(os.path.join(args[2], "log", LOG_PIPELINE_FILE_NAME), logger.level)
        _logger = logger
        _logger.addHandler(_handler)
        try:
            result = func(self, *args, **kwargs)
            return result
        except Exception as e:
            _logger.exception(e)
            raise e
        finally:
            _logger.removeHandler(_handler)
            _handler.close()
    return wrapper


RESOURCE_MESSAGE_FORMAT = "".join(["\033[93m",
                                   "\n------------------------------------ %s < CPU/MEMORY/SUMMARY> Info ------------------------------------",
                                   "\n%s",
                                   "\n%s",
                                   "\n%s",
                                   "\033[0m"])


def profile_resource(func):
    """ cpu/memory profiling decorator

    Args:
        func    (function): original function

    Returns:
        wrapper (function): wrapped function

    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        if not settings.experimental_plan.control.check_resource:
            return func(*args, **kwargs)
        pid = os.getpid()
        ppid = psutil.Process(pid)
        cpu_usage_start = ppid.cpu_percent(interval=None)  # 단순 cpu 사용률
        mem_usage = ResourceProfile(ppid, cpu_usage_start)
        thread = Thread(target=mem_usage.run, daemon=True)
        thread.start()
        result = func(*args, **kwargs)
        mem_usage.enable = False
        cpu, mem = mem_usage.info()
        msg_cpu = "- CPU (min/max/avg) : {:5.1f}% / {:5.1f}% / {:5.1f}%".format(*cpu)
        msg_mem = "- MEM (min/max/avg) : {} / {} / {}".format(*mem)
        pipes = []
        context = args[1]
        stage_name = args[2]
        for pipe_name in context[stage_name][RUN_PIPELINE_NAME]:
            pipes.append(f"{stage_name} - {pipe_name:<15} : "
                         f"Elapsed time ({(context[stage_name][pipe_name]['finishAt'] - context[stage_name][pipe_name]['startAt']).total_seconds():8.3f}) "
                         f"[{context[stage_name][pipe_name]['finishAt'].strftime('%Y-%m-%d %H:%M:%S.%f')}"
                         f" - {context[stage_name][pipe_name]['startAt'].strftime('%Y-%m-%d %H:%M:%S.%f')}]")
        logger.debug(RESOURCE_MESSAGE_FORMAT, stage_name, msg_cpu, msg_mem, "\n".join(pipes))
        return result

    return wrapper


def save_summary(solution_metadata_version: str, file_path: str, result="", score="", note="", probability={}):
    """ Save train_summary.yaml (when summary is also conducted during train) or inference_summary.yaml.
        e.g. self.asset.save_summary(result='OK', score=0.613, note='alo.csv', probability={'OK':0.715, 'NG':0.135, 'NG1':0.15}

    Args:
        solution_metadata_version (str) : version of solution_metadata
        file_path   (str): Path where the file will be saved
        result      (str): Inference result summarized info. (length limit: 25)
        score       (float): model performance score to be used for model retraining (0 ~ 1.0)
        note        (str): optional & additional info. for inference result (length limit: 100) (optional)
        probability (dict): probability per class prediction if the solution is classification problem.  (optional)
                            e.g. {'OK': 0.6, 'NG':0.4}

    Returns:
        summaray_data   (dict): data tobe saved in summary yaml

    """
    result_len_limit = 32
    note_len_limit = 128
    if result is None:
        result = None
    elif not isinstance(result, str) or len(result) > result_len_limit:  # check result length limit 12
        logger.warning("The summary['result'] value must be a str type and the length must be less than %d characters. "
                       "Any characters exceeding the string length are ignored.", result_len_limit)
        result = str(result)[:result_len_limit]
    if score is None:
        score = None
    elif not type(score) in (int, float) or not 0 <= score <= 1.0:  # check score range within 0 ~ 1.0
        logger.warning("The summary['score'] value must be python float or int. Also, the value must be between 0.0 and 1.0. "
                       "Your current score value: %s", score)
        score = 0.0
    if not isinstance(note, str) or len(note) > note_len_limit:  # check note length limit 100
        logger.warning("The summary['note'] value must be a str type and the length must be less than %d characters. "
                       "Any characters exceeding the string length are ignored.", note_len_limit)
        note = str(note)[:note_len_limit]
    if (probability is not None) and (not isinstance(probability, dict)):  # check probability type (dict)
        raise AloErrors['ALO-SMM-001']("The 'probability' property of the pipeline function result summary must be of dict data type.",
                                       doc={'keyType': type(probability).__name__})
    if len(probability.keys()) > 0:  # check type - probability key: string,value: float or int
        prob_sum = 0
        for k, v in probability.items():
            if not isinstance(k, str) or type(v) not in (float, int):
                raise AloErrors['ALO-SMM-002']("The 'probability' dict item in summary can only store key(str):value(int or float) data types.",
                                               doc={'key':k, 'keyType': type(k).__name__, 'value': v, 'valueType': type(v).__name__})
            prob_sum += v
        if round(prob_sum) != 1:  # check probability values sum = 1
            raise AloErrors['ALO-SMM-003']("The sum of the summary probability values must equal 1.0.",
                                           doc={'values': list(probability.values()), 'sumValues':prob_sum})
    else:
        pass
        # FIXME e.g. 0.50001, 0.49999 case?

    # FIXME it is necessary to check whether the sum of the user-entered dict is 1, anticipating a floating-point error
    def make_addup_1(prob):
        # Process the probabilities to sum up to 1, displaying up to two decimal places
        max_value_key = max(prob, key=prob.get)
        proc_prob_dict = dict()
        for k, v in prob.items():
            if k == max_value_key:
                proc_prob_dict[k] = 0
                continue
            proc_prob_dict[k] = round(v, 2)
        proc_prob_dict[max_value_key] = round(1 - sum(proc_prob_dict.values()), 2)
        return proc_prob_dict

    if (probability is not None) and (probability != {}):
        probability = make_addup_1(probability)
    else:
        probability = {}

    if score is not None:
        summary_data = {
            'result': result,
            'score': round(score, 2),
            'date': datetime.now(timezone('UTC')).strftime('%Y-%m-%d %H:%M:%S'),
            'note': note,
            'probability': probability,
            'version': solution_metadata_version
        }
    else:
        summary_data = {
            'result': result,
            'score': score,
            'date': datetime.now(timezone('UTC')).strftime('%Y-%m-%d %H:%M:%S'),
            'note': note,
            'probability': probability,
            'version': solution_metadata_version
        }

    with open(file_path, 'w') as file:
        yaml.dump(summary_data, file, default_flow_style=False)
    logger.debug("[SUMMARY] Successfully saved summary yaml : %s", file_path)

    return summary_data


class WorkspaceDict(dict):
    """workspace 이하의 파일 관리를 위한 객체

    Keys:
        - workspace (str): 작업 경로

    """
    def __init__(self, workspace: str):
        """
        작업 경로 폴더를 생성 후 workspace를 등록합니다.
        Args:
            workspace: 작업 기본 경로
        """
        super().__init__()
        self.update_workspace(workspace)
        Path(self['workspace']).mkdir(exist_ok=True)

    def update_workspace(self, value):
        super().__setitem__('workspace', value)

    def __setitem__(self, key, value):
        """ 신규 키-값 을 추가합니다.

        Args:
            key: 키
            value: 값
        Raises:
            - ALO-CTX-001 : key 데이터 유형이 문자가 아닌 경우
            - ALO-CTX-002 : 시스템 예약어 workspace 속성을 수정하려는 경우
            - ALO-CTX-003 : key 명명 규칙 위반 시

        """
        if not isinstance(key, str):
            raise AloErrors['ALO-CTX-001'](f'"{str(key)}({type(key)})" is not allowed as a key value. The key value must be of type string.',
                                           doc={'key': key, 'type': type(key).__name__})
        if key == 'workspace' and key in self:
            raise AloErrors['ALO-CTX-002']('The word "workspace" is a system reserved word and cannot be update. Change the "workspace" keyword to another name.',
                                           doc={'key': 'workspace'})

        # if not re.match(r'^[ㄱ-힣a-zA-Z0-9 _\-,\.\/]+$', key):
        #     raise AloErrors['ALO-CTX-003']('The key contains an invalid string.',
        #                                    doc={'invalidKey': key, 'allowedRegex': r'^[ㄱ-힣a-zA-Z0-9 _\-,\.\/]+$'})

        if not re.match(r'^[ㄱ-힣a-zA-Z0-9 _\-,\.\/]+$', key):
            logger.warning('The key contains a special symbols.')
                                           #doc={'invalidKey': key, 'allowedRegex': r'^[ㄱ-힣a-zA-Z0-9 _\-,\.\/]+$'})

        super().__setitem__(key, value)

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return self['workspace']

    def __radd__(self, other):
        return other + str(self)

    def __add__(self, other):
        return str(self) + other


class ArtifactModel(WorkspaceDict):
    """ 학습/추론 작업에 의한 산출물(파일)을 관리하기 위한 객체

    """
    def __init__(self, stage_name: str, stage_workspace: str):
        super().__init__(f"{stage_workspace}{os.sep}output")
        self.__stage_name = stage_name

    def __setitem__(self, key, value):
        super().__setitem__(key, value)

    def validate(self):
        """ 산출물 파일들에 대한 검증

        Raises:
            ALO-ART-001 : 산출물 파일이 없는 경우
            ALO-ART-002 : 2개를 초과하여 파일을 저장한 경우
            ALO-ART-003 : 허용 가능한 파일 확장자가 아닌 경우.
                저장 가능 확장자(csv, jpg, jpeg, png, svg)
        """
        files = [file.lower() for file in os.listdir(self['workspace']) if os.path.isfile(os.path.join(self['workspace'], file))]
        if self.__stage_name == INFERENCE and not files:
            raise AloErrors['ALO-ART-001']("The output file could not be found. In the inference phase, you must create one or two files under the path `pipeline['artifact']['workspace']`.",
                                           doc={"stage": self.__stage_name})
        if not files:
            return
        if len(files) > 2:
            raise AloErrors['ALO-ART-002']('You have to save inference output file. The number of output files must be 1 or 2.',
                                           doc={"stage": self.__stage_name, "files": files})
        if not all([file.lower().endswith((".csv", ".jpg", ".jpeg", ".png", ".svg")) for file in files]):
            raise AloErrors['ALO-ART-003']('output file extension must be one of ["csv", "jpg", "jpeg", "png", "svg"].',
                                           doc={"stage": self.__stage_name, "files": files})

class ExtraOutputModel(WorkspaceDict):
    """ 추가 출력물을 관리하기 위한 객체 """
    def __init__(self, stage_workspace: str):
        super().__init__(f"{stage_workspace}{os.sep}extra_output")
        Path(self['workspace']).mkdir(parents=True, exist_ok=True)

    def __setitem__(self, key, value):
        super().__setitem__(key, value)

class TrainModel(WorkspaceDict):
    """학습의 모델과 관련된 파일 및 workspace 경로 정보를 관리하기 위한 객체

    workspace이하의 pkl(pickle) 파일을 메모리로 로드 후 해당 객체를 전달한다.

    Keys:
        - workspace (str): 작업 경로

    Raises:
        AloErrors: ALO-PIP-007 (저장된 모델이 없는 경우 예외 발생)

    Examples:
        >>> context['model']['workspace']
           /home/alo/test
        >>> str(context['model'])  # context['model']['workspace'] 와 동일한 값을 리턴함
           /home/alo/test
        >>> context['model']['titanic_model']
           RandomForestClassifier 객체를 리턴함
        >>> context['model']['titanic_model'] = RandomForestClassifier(n_estimators=n_estimators, max_depth=5, random_state=1)
           RandomForestClassifier 객체를 titanic_model 이름으로 파일로 저장됨

    """
    MODEL_FILE_NAME_FORMAT = "{0}.pkl"

    def __init__(self, workspace: str):
        super().__init__(workspace)

    def __getitem__(self, key):
        if key in self:
            return super().get(key)

        file = f"{super().get('workspace')}{os.sep}{self.MODEL_FILE_NAME_FORMAT.format(key)}"
        if not os.path.isfile(file):
            raise AloErrors['ALO-MDL-001'](f'"{key}" cannot be found in context["model"]. Check model name or model_uri',
                                           doc={"key": key, "file": file})

        with open(file, 'rb') as f:
            try:
                self[key] = pickle.load(f)
            except Exception as e:
                raise AloErrors['ALO-MDL-002'](f'Failed to unpickle : {key}',
                                               doc={"key": key, "file": file}) from e
        logger.debug('[MODEL] Load context["model"]["%s"] : %s', key, file)
        return super().get(key)

    def __setitem__(self, key, value):
        try:
            with open(os.path.join(self['workspace'], self.MODEL_FILE_NAME_FORMAT.format(key)), 'wb') as f:
                pickle.dump(value, f)
                super().__setitem__(key, value)
                logger.debug('[MODEL] save context["model"]["%s"] : %s/%s', key, self['workspace'], self.MODEL_FILE_NAME_FORMAT.format(key))
        except Exception as e:
            raise AloErrors['ALO-MDL-003'](f'Failed to save model "{key}".',
                                           doc={"key": key,
                                                "file": os.path.join(self['workspace'], self.MODEL_FILE_NAME_FORMAT.format(key))}) from e

    def __deepcopy__(self, memodict={}):
        return self

    def validate(self, phase: str = None):
        workspace = self['workspace']
        files = [f.lower() for f in os.listdir(workspace) if os.path.isfile(os.path.join(workspace, f))]
        directories = [d.lower() for d in os.listdir(workspace) if os.path.isdir(os.path.join(workspace, d))]

        if phase == TRAIN and not (files or directories):
            raise AloErrors['ALO-MDL-004']("When training, you must save at least one model/config file or directory.",
                                        doc={
                                            "phase": phase,
                                            "files": files,
                                            "directories": directories
                                        })


class Dataset(WorkspaceDict):
    """
    train/inference 시 입력 데이터셋으로 활용할 파일들에 대한 폴더 및 파일 목록

    Keys:
        - workspace (str): 작업 경로
    """

    def __init__(self, stage_name: str, stage_workspace: str):
        super().__init__(f"{stage_workspace}{os.sep}dataset")
        self.__stage_name = stage_name

    def __getitem__(self, key):
        if key not in self:
            raise AloErrors['ALO-DTS-001'](f'"{key}" not exists in pipeline["dataset"]. Check dataset in {self.__stage_name}',
                                           doc={"stage": self.__stage_name, "key": key})
        if key == 'workspace':
            return super().__getitem__(key)
        if not os.path.isfile(f'{self["workspace"]}{os.sep}{key}'):
            raise AloErrors['ALO-DTS-002'](f'"{key}" file cannot be found in {self["workspace"]}.',
                                           doc={"stage": self.__stage_name, "key": key, "file": f'{self["workspace"]}{os.sep}{key}'})
        return super().__getitem__(key)

    def add(self, files: list):
        prefix_len = len(f'{self["workspace"]}{os.sep}')
        for file in files:
            self[file[prefix_len:]] = file

    def __deepcopy__(self, memodict={}):
        return self


class Context(WorkspaceDict):
    """ ALO 수행과 관련된 환경 정보를 담고 있는 객체

    Keys:
        - workspace (str): 작업 경로
        - startAt (datetime): 시작 시각
        - finishAt (datetime): 종료 시각(완료된 후 key 값이 추가됨)

    """
    def __init__(self):
        start_at = datetime.now()
        super().__init__(f"{settings.history_path}{os.sep}{start_at.strftime(HISTORY_FOLDER_FORMAT)}")
        self['startAt'] = start_at
        Path(self['workspace']).mkdir(parents=True, exist_ok=True)
        self['id'] = str(uuid.uuid4())
        self['name'] = settings.name
        self['version'] = settings.version
        self['host'] = settings.host
        self['logger'] = logger
        self['logging'] = {
            'name': 'alo',
            'level': settings.log_level,
        }
        self['model'] = TrainModel(settings.model_artifacts_path)
        self['external'] = WorkspaceDict(f"{self['workspace']}{os.sep}external")
        self['stage'] = None
        self['solution_metadata_version'] = settings.experimental_plan.solution.version

    def __enter__(self):
        return self

    def __getitem__(self, key):
        if key not in self and key in MODES:
            stage_ws = f"{self['workspace']}{os.sep}{key}"
            Path(stage_ws).mkdir(parents=True, exist_ok=True)
            self[key] = {
                'name': key,
                'workspace': stage_ws,
                'logger': self['logger'],
                'model': self['model'],
                'external': self['external'],
                'dataset': Dataset(key, stage_ws),
                ARTIFACT: ArtifactModel(key, stage_ws),
                'extra_output': ExtraOutputModel(stage_ws),  # 추가된 부분
                RUN_PIPELINE_NAME: []
            }
            Path(os.path.join(stage_ws, "score")).mkdir()
        return super().__getitem__(key)

    def __create_result_info(self):
        def order_dict(dictionary: dict):
            return {k: order_dict(v) if isinstance(v, dict) else v for k, v in sorted(dictionary.items())}

        shutil.copy2(settings.experimental_plan.uri, os.path.join(self['workspace'], EXP_FILE_NAME))
        with open(settings.experimental_plan.uri, 'rb') as plan_f, open(os.path.join(self['workspace'], RESULT_INFO_FILE), 'w') as json_f:
            info = {
                'start_time': self['startAt'].isoformat() if self.get('startAt') else None,
                'end_time': self['finishAt'].isoformat() if self.get('finishAt') else None,
                EXP_FILE_NAME: {
                    'modify_date': datetime.fromtimestamp(os.path.getmtime(settings.experimental_plan.uri)).isoformat()
                },
                **{mode: {'start_time': self[mode]['startAt'].isoformat() if self[mode].get('startAt') else None,
                          'end_time': self[mode]['finishAt'].isoformat() if self[mode].get('finishAt') else None,
                          'argument': {pipe: order_dict(self[mode][pipe].get('argument', {})) for pipe in self[mode][RUN_PIPELINE_NAME]}
                          }
                   for mode in MODES if mode in self}
            }
            json.dump(info, json_f)

    def __exit__(self, exc_type, exc_val, exc_tb):
        latest = 'latest'
        if isinstance(exc_val, Exception):
            logger.error("An error occurred: %s", exc_val)
            rename_ws = f"{self['workspace']}-error"
            os.rename(self['workspace'], rename_ws)
            self.update_workspace(rename_ws)
            latest = f'{latest}-error'
            logger.error("Please check the detailed log: %s", rename_ws)
            for stage_name in MODES:
                if stage_name in self:
                    self[stage_name]['workspace'] = f"{self['workspace']}{os.sep}{stage_name}"
        latest_link = os.path.join(settings.history_path, latest)
        if os.path.islink(latest_link):
            os.unlink(latest_link)
        os.symlink(self['workspace'], latest_link)
        self['finishAt'] = datetime.now()
        if settings.experimental_plan.uri:
            self.__create_result_info()

        logger.info('[CONTEXT] Total elapsed second : %.2f', self.elapsed_seconds)
        self.retain_history()

    def retain_history(self):
        paths = settings.experimental_plan.control.backup.retain(settings.history_path)
        logger.debug("[HISTORY] remove old directory : %s", paths)

    @property
    def elapsed_seconds(self):
        return ((self['finishAt'] if self.get('finishAt') else datetime.now()) - self['startAt']).total_seconds()

    def summary(self, phase_name: str):
        phase = self.get(phase_name)
        if not phase:
            return None
        summaries = {}
        for pipe_name in phase.get(RUN_PIPELINE_NAME, []):
            summary = phase.get(pipe_name, {}).get('summary', None)
            if summary:
                summaries[pipe_name] = summary
        if len(summaries) == 1:
            for _, v in summaries.items():
                return v
        return summaries


def _v1_convert_sol_args(stage_name, _args):
    """ - Delete any args in the selected user parameters that have empty values.
        - Convert string type comma splits into a list.

    Args:
        _args   (dict): args tobe converted

    Returns:
        _args   (dict): converted args
    """
    # TODO Should we check the types of user parameters to ensure all selected_user_parameters types are validated?
    if not isinstance(_args, dict):
        raise AloErrors['ALO-PIP-009'](f"selected_user_parameters args. in solution_medata must have << dict >> type : {_args}", pipeline=stage_name)
    if not _args:
        return _args
    # when a multi-selection comes in empty, the key is still sent \
    # e.g. args : { "key" : [] }
    _args_copy = deepcopy(_args)
    for k, v in _args_copy.items():
        # single(multi) selection
        # FIXME Although a dict type might not exist, just in case... \
        # (perhaps if a dict needs to be represented as a str, it might be possible?)
        if isinstance(v, list) or isinstance(v, dict):
            if len(v) == 0:
                del _args[k]
        elif isinstance(v, str):
            if (v is None) or (v == ""):
                del _args[k]
            else:
                # 'a, b' --> ['a', 'b']
                converted_string = [i.strip() for i in v.split(',')]
                if len(converted_string) == 1:
                    # ['a'] --> 'a'
                    _args[k] = converted_string[0]
                elif len(converted_string) > 1:
                    # ['a', 'b']
                    _args[k] = converted_string
                    # int, float
        else:
            if v is None:
                del _args[k]
    return _args


class Computing(metaclass=ABCMeta):
    """ 학습/추론 기능 구현을 위한 추상클래스

    """

    def __init__(self):
        self.experimental_plan = None
        self.solution_metadata = None
        print_copyright(COPYRIGHT)
        self.reload()


    def init(self):
        settings.update()
        self.experimental_plan = settings.experimental_plan
        self.solution_metadata = settings.solution_metadata
        if not self.experimental_plan:
            raise AloErrors['ALO-INI-000']('experimental_plan.yaml information is missing.')

    def install(self):
        source_path = self.checkout_git()
        self.install_pip(source_path)
        self.load_module()

    def reload(self):
        """ 환경 설정 정보 및 library 재설정
        """
        self.init()
        self.install()
        self.show_version()

    def run(self):
        try:
            self.solve()
        except Exception as e:
            error = e if isinstance(e, AloError) else AloError(str(e))
            logger.exception(error)
            raise error

    def show_version(self):
        logger.info("\033[96m\n=========================================== Info ==========================================="
                    f"\n- Time (UTC)        : {datetime.now(timezone('UTC')).strftime('%Y-%m-%d %H:%M:%S')}"
                    f"\n- Alo               : {__version__}"
                    f"\n- Solution Name     : {self.experimental_plan.name}"
                    f"\n- Solution Version  : {self.experimental_plan.version}"
                    f"\n- Solution Plan     : {self.experimental_plan.uri}"
                    f"\n- Solution Meta     : {self.solution_metadata.uri if self.solution_metadata else ''}"
                    f"\n- Home Directory    : {settings.home}"
                    "\n============================================================================================\033[0m")

    def load_module(self):
        if self.experimental_plan.solution:
            self.experimental_plan.solution.update_pipeline()

    @abstractmethod
    def solve(self):
        pass

    def exec_stage(self, context, stage_name):
        context['stage'] = stage_name
        self.stage(context, stage_name, f'{context["workspace"]}/{stage_name}')
        context['model'].validate(stage_name)
        context[stage_name][ARTIFACT].validate()

    @add_logger_handler
    @log_start_finish(logger, "{}", highlight=True, args_indexes=[1])
    @profile_resource
    def stage(self, context, stage_name, stage_workspace):
        stage = getattr(self.experimental_plan.solution, stage_name)
        if not stage:
            logger.debug("[PIPELINE] Empty %s info. Skip %s", stage_name, stage_name)
            return
        for pipe_name, function in stage.pipeline.items():
            logger.debug("[PIPELINE] %10s : %15s - %s.%s", stage_name, pipe_name, function.def_.__module__, function.def_.__name__)

        pipeline = context[stage_name]
        dataset_files = stage.get_dataset(pipeline['dataset']['workspace'])
        pipeline['dataset'].add(dataset_files)

        logger.debug('[PIPELINE] List of imported dataset:\n%s', "\n".join(dataset_files))
        extract_file(dataset_files, pipeline['dataset']['workspace'])
        model_files = stage.get_model(context['model']['workspace'])
        logger.debug('[PIPELINE] List of imported model:\n%s', "\n".join(model_files))
        extract_file(model_files, context['model']['workspace'])
        pipeline['startAt'] = datetime.now()

        for i, (pipe_name, function) in enumerate(stage.pipeline.items()):
        # for pipe_name, function in stage.pipeline.items():
            if settings.mode_pipeline and pipe_name not in settings.mode_pipeline:
                logger.warning("[PIPELINE] Skip solution.%s.%s : --mode_pipeline %s", stage_name, pipe_name, settings.mode_pipeline)
                continue
            pipeline[pipe_name] = {
                'startAt': datetime.now(),
                'workspace': pipeline['workspace'],
                ARTIFACT: pipeline[ARTIFACT],
            }
            # Get data files
            is_last_iteration = (i == len(stage.pipeline) - 1)
            self.pipeline(context, pipeline, pipeline[pipe_name], pipe_name, function, is_last_iteration)
            pipeline[pipe_name]['finishAt'] = datetime.now()
            pipeline[RUN_PIPELINE_NAME].append(pipe_name)
        pipeline['finishAt'] = datetime.now()

        return pipeline

    def __get_func_args(self, context, pipeline, handler, handler_kwargs):
        kwargs = handler_kwargs
        if handler.__defaults__ is None:
            kwargs = {}

        args_cnt = handler.__code__.co_argcount - (0 if handler.__defaults__ is None else len(handler.__defaults__))
        if args_cnt == 0:
            return [], kwargs
        elif args_cnt == 1:
            return [pipeline], kwargs
        elif args_cnt == 2:
            return [context, pipeline], kwargs
        else:
            raise AloErrors["ALO-USR-002"]("Invalid arguments of function handler",
                                           doc={"file": inspect.getfile(handler),
                                                "function": f"{handler.__name__}()",
                                                "message": "The number of positional arguments for a function handler cannot exceed two."})

    @log_start_finish(logger, "{} pipline", highlight=False, args_indexes=[3])
    def pipeline(self, context: dict, pipeline: OrderedDict, pipe: dict, name: str, function, is_last_iteration: bool) -> dict:
        func_kwargs = function.get_argument()
        # before
        pipeline[name]['argument'] = func_kwargs
        clone_context = deepcopy(context)
        logger_fn = logger.error
        try:
            args, kwargs = self.__get_func_args(clone_context, clone_context[context['stage']], function.def_, func_kwargs)
            result = function.def_(*args, **kwargs)
            is_default = False
            # 함수 이름이 train 또는 inference인 경우 summary 키 검증
            # if 'train' in function.def_.__name__ or 'inference' in function.def_.__name__: # train inf가 모두 포함되게 수정
            def make_summary(result, name):
                is_default = False
                if not isinstance(result, dict):
                    logger.warning(
                        "[PIPELINE] '%s' function should return a dict with 'summary' key. Creating default summary.",
                        name # function.def_.__name__
                    )
                    result = {}

                if 'summary' not in result:
                    logger.warning(
                        "[PIPELINE] Missing 'summary' key in '%s' function result. Creating default summary.",
                        name # function.def_.__name__
                    )
                    result['summary'] = {}

                required_summary_keys = {'result', 'note', 'score'}
                missing_keys = required_summary_keys - set(result['summary'].keys())

                if missing_keys:
                    logger.warning(
                        "[PIPELINE] Missing required keys %s in summary dict. Adding default values.",
                        list(missing_keys)
                    )

                    # 누락된 키에 대한 기본값 설정
                    default_values = {
                        'result': None,
                        'note': None,
                        'score': None
                    }

                    # 누락된 키만 기본값으로 채움
                    for key in missing_keys:
                        result['summary'][key] = default_values[key]
                        is_default = True

                return result, is_default

            if self.experimental_plan.api is None:
                result, is_default = make_summary(result, function.def_.__name__)

            context['model'].update(clone_context['model'])
            context['external'].update(clone_context['external'])
            pipeline[name]['result'] = result
            # after
            summary = self.save_output(context, pipeline['name'], pipe, result, is_last_iteration, is_default)
            logger_fn = logger.info
        except AloError as e:
            raise e
        except Exception as e:
            raise AloErrors["ALO-USR-001"](str(e), doc={"file": inspect.getfile(function.def_), "function": f"{function.def_.__name__}()", "message": str(e)}) from e
        finally:
            logger_fn(
                "[PIPELINE] function call info\n"
                "***************************** Invoke Pipline Function *****************************\n"
                "* Target File             : %s\n"
                "* function[name]          : %s\n"
                "* function[name].def      : %s.%s\n"
                "* function[name].argument : %s\n"
                "* summary                 : %s\n"
                "***********************************************************************************",
                inspect.getfile(function.def_), name, function.def_.__module__, function.def_.__name__, func_kwargs, pipe.get('summary', ''))

    def checkout_git(self):
        try:
            if self.experimental_plan.solution is None or self.experimental_plan.solution.git is None:
                logger.info('[GIT] "git" property is not set.')
                return
            name = self.experimental_plan.solution.git.url.path.split('/')[-1].split('.')[0]
            path = f"{settings.workspace}/{name}"
            self.experimental_plan.solution.git.checkout(path)
            logger.debug("[GIT] checkout : %s -> %s", self.experimental_plan.solution.git.url, path)
            return path
        except Exception as e:
            raise AloErrors["ALO-PIP-001"](str(e)) from e

    def install_pip(self, source_path: str):
        try:
            if self.experimental_plan.solution is None or not self.experimental_plan.solution.pip:
                return
            if source_path is None:
                source_path = os.path.dirname(self.experimental_plan.uri)
            req_file = os.path.join(source_path, 'requirements.txt')
            if self.experimental_plan.solution.pip.requirements is True and not os.path.exists(req_file):
                raise AloErrors["ALO-INI-003"](req_file)

            install_packages = []
            if self.experimental_plan.solution.pip.requirements is True:
                install_packages.append(f"-r {req_file}")
            elif isinstance(self.experimental_plan.solution.pip.requirements, list):
                for req in self.experimental_plan.solution.pip.requirements:
                    if req.endswith('.txt'):
                        req_file = os.path.join(os.path.dirname(self.experimental_plan.uri), req)
                        if not os.path.exists(req_file):
                            raise AloErrors["ALO-INI-003"](req_file)
                        req = f"-r {req_file}"
                    install_packages.append(req)
            else:
                logger.debug("[PIP] Skip pip install")
                return

            installed_packages = []
            self.experimental_plan.solution.pip.convert_req_to_list(self.experimental_plan.solution.pip.requirements)
            for package in install_packages:
                try:
                    exists_package = pkg_resources.get_distribution(package)
                    installed_packages.append(package)
                    logger.debug("[PIP] %s already installed: %s", package, exists_package)
                except Exception:
                    logger.debug("[PIP] Start installing package - %s", package)
                    self.experimental_plan.solution.pip.install(package)
                    installed_packages.append(package)
        except Exception as e:
            raise AloErrors['ALO-INI-002'](str(e)) from e

    def save_output(self, context: dict, pipe_name: str, pipe: dict, output, is_last_iteration: bool, is_default: bool):
        if output is None:
            return
        if isinstance(output, dict):
            summary_file_path = os.path.join(pipe['workspace'], "score", f"{pipe_name}_summary.yaml")
            if output.get('summary'):
                if is_last_iteration or is_default: # 마지막 iter이거나 default인 경우 작성
                    if not os.path.exists(summary_file_path): # 이전에 작성한 summary가 있으면 작성하지 않음
                        summary_data = save_summary(context['solution_metadata_version'], summary_file_path, **output.get('summary'))
                if not is_default: # default 가 아닌 경우는 무조건 작성
                    summary_data = save_summary(context['solution_metadata_version'], summary_file_path, **output.get('summary'))
                    pipe['summary'] = summary_data

    def artifact(self, context, stage_name):
        stage = getattr(self.experimental_plan.solution, stage_name)
        pipeline = context[stage_name]
        put_files = []

        def filter_compress(info):  # dataset 폴더는 압축 대상에서 제외
            if re.search(r'^[^\/]+\/dataset(\/|$)', info if isinstance(info, str) else info.name):
                return None
            else:
                return info

        if stage_name == TRAIN:
            put_files.extend(stage.put_data(settings.model_artifacts_path, "model"))  # model.tar.gz
        shutil.copy2(os.path.join(settings.log_path, LOG_PROCESS_FILE_NAME), os.path.join(pipeline['workspace'], 'log'))
        put_files.extend(stage.put_data(os.path.join(pipeline['workspace'], "log", LOG_PIPELINE_FILE_NAME), LOG_PIPELINE_FILE_NAME))  # pipeline.log
        put_files.extend(stage.put_data(os.path.join(settings.log_path, LOG_PROCESS_FILE_NAME), LOG_PROCESS_FILE_NAME))  # process.log
        put_files.extend(stage.put_data(pipeline['workspace'], f"{stage_name}_artifacts", filter=filter_compress,
                                        compress_type=None if stage_name == TRAIN else 'zip'))  # train targ.gz, inference zip
        logger.debug('[ARTIFACT] List of artifacts :\n%s', "\n".join(put_files))
        logger.debug("[ARTIFACT] Success save to : %s", stage.artifact_uri)

    def update_experimental_plan(self, stage_name: str):
        # todo overwrite plan.
        # todo Hardcoded.
        # todo Need more generic convention method.
        if not self.solution_metadata:
            logger.info("[YAML] Skip update experimental_plan property:  Empty solution_metadata.")
            return
        source = self.solution_metadata.get_pipeline(stage_name)
        if not source:
            logger.info("[YAML] Skip update experimental_plan property: Empty %s pipeline information in solution_metadata.", stage_name)
            return
        target = getattr(self.experimental_plan.solution, stage_name)
        if not target:
            logger.info("[YAML] Skip update experimental_plan property: Empty %s pipeline information in experimental_plan.", stage_name)
            return

        # update uri, aws credential
        for uri in ['dataset_uri', 'model_uri', 'artifact_uri']:
            source_uri = getattr(source, uri, None)
            if not source_uri:
                setattr(target, uri, [])  # None 인 경우 미동작으로 설정
                continue
            setattr(target, uri, source_uri)
            update_storage_credential(self.experimental_plan.solution.credential, target)

        # update selected_user_parameters
        for func_name, function in target.pipeline.items():
            if not source.parameters or not source.parameters.get_type_args("selected_user_parameters", func_name):
                logger.debug("[YAML] Skip update: pipeline.paramters.[%s] not define in solution_metadata.", func_name)
                continue
            sol_args = _v1_convert_sol_args(stage_name, source.parameters.get_type_args("selected_user_parameters", func_name))
            if sol_args:
                function.update(sol_args)  # 정의되지 않은 속성의 경우 기존 값을 그대로 사용하게 됨 v1 로직 유지

        self.experimental_plan.solution.version = self.solution_metadata.metadata_version


    def history(self, data_id="", param_id="", code_id="", parameter_steps=[], type=MODES, head: int = None, tail: int = None, show_table=False):
        """ Deliver the experiment results stored in history as a table,
            allowing for solution registration by history id.
            After verifying the identity between experimental_plan.yaml in the
            history folder and each id, create a table.

        Args:
            data_id         (str): data id
            param_id        (str): parameters id
            code_id         (str): source code id
            parameter_steps (list): decide which step's parameters to display when creating a table
            type            (str): train or inference (default: [train, inference])
            head            (int): output the first part of history
            tail            (int): output the first part of history

        Returns: -

        """
        if self.experimental_plan is None:
            self.init()
        scores = []

        def make_score(event: str, mode: str, status: str, result_info: dict, summary: dict):
            score = {
                'id': event,
                'status': status,
                'type': mode,
                **result_info.get(mode, {}),
                **{i: summary.get(i, None) for i in ['pipeline_name', 'score', 'result', 'note', 'probability', 'version']},
                'checksum': {**summary.get('checksum', {}),
                             EXP_FILE_NAME: result_info.get('checksum', None)}
            }
            scores.append(score)

        pipeline_type = type if isinstance(type, list) else [type]
        dirs = [(folder, os.path.join(settings.history_path, folder),) for folder in os.listdir(settings.history_path) if not folder.startswith('latest')]
        dirs = sorted(dirs, key=lambda x: x[0])
        dir_size = len(dirs)
        if head:
            dirs = dirs[:head]
        if tail:
            dirs = dirs[-tail:]
        for folder, folder_path in dirs:
            name_group = HISTORY_PATTERN.search(folder)
            if not name_group:
                continue

            result_info_file = os.path.join(folder_path, RESULT_INFO_FILE)
            if not os.path.exists(result_info_file):
                scores.append({'id': name_group[1], 'status': 'error' if name_group[2] else 'success'})
                continue
            with open(result_info_file, 'r') as f, open(os.path.join(folder_path, EXP_FILE_NAME), 'r') as plan_f:
                try:
                    result_info = json.load(f)
                except Exception as e:
                    result_info = {}
                result_info['checksum'] = hashlib.md5(plan_f.read().encode()).hexdigest()[:8]

            for mode in [pipe for pipe in pipeline_type if pipe in MODES]:
                if mode not in result_info:
                    continue
                summary_files = [(os.path.join(folder_path, mode, "score", temp), temp.replace('_summary.yaml', ''))
                                 for temp in os.listdir(os.path.join(folder_path, mode, "score")) if not folder.endswith('_summary.yaml')]
                for file_path, pipeline_name in summary_files:
                    summary = {'pipeline_name': pipeline_name}
                    if os.path.isfile(file_path):
                        try:
                            with open(file_path, 'r') as f:
                                summary = yaml.safe_load(f)
                                summary['pipeline_name'] = pipeline_name
                        except Exception as e:
                            logger.warning("An error occurred while reading the file : %s", file_path)
                    summary['checksum'] = {'argument': hashlib.md5(json.dumps(result_info.get(mode, {}).get('argument', {})).encode()).hexdigest()[:8],
                                           'dataset': {}}
                    len_dir_path = len(os.path.join(folder_path, mode, 'dataset'))
                    for root, _, files in os.walk(os.path.join(folder_path, mode, 'dataset')):
                        for file in files:
                            file_path = os.path.join(root, file)
                            with open(file_path, 'r') as f:
                                summary['checksum']['dataset'][file_path[len_dir_path + 1:]] = hashlib.md5(f.read().encode()).hexdigest()[:8]
                    make_score(name_group[1], mode, 'error' if name_group[2] else 'success', result_info, summary)

        if show_table:
            print_table(scores, **(show_table if isinstance(show_table, dict) else {}))

        return scores


class OneTime(Computing, metaclass=ABCMeta):
    """ 학습/추론 과정을 1회만 수행하는 클래스

    프로그램 종료됨

    """

    def solve(self, phases: list = None):
        if phases is None:
            modes = MODES if settings.mode is None else [settings.mode]
        else:
            modes = phases
        with Context() as context:
            for name in MODES:
                if name not in modes:
                    logger.info("Skip %s()", name)
                    continue
                if not getattr(self.experimental_plan.solution, name):
                    logger.info("Skip solution.%s : Not define in experimental_plan.yaml", name)
                    continue

                try:
                    if name == TRAIN:
                        self.init_train()
                    self.update_experimental_plan(name)
                    self.exec_stage(context, name)
                except Exception as e:
                    self.artifact(context, name)
                    raise e
                else:
                    self.artifact(context, name)

    def init_train(self):
        files = glob.glob(os.path.join(settings.model_artifacts_path, "*"))
        for f in files:
            os.remove(f) if os.path.isfile(f) else shutil.rmtree(f)

    def __update_func_kwargs(self, stage_name: str, argument: dict):
        stage = getattr(self.experimental_plan.solution, stage_name, None)
        if stage is None or argument is None:
            return
        if not isinstance(argument, dict):
            raise ValueError("""train/inference function argument must be dict type. Ex)
{
    'preprocess': {
        'x_columns': [ 'Pclass', 'Sex', 'SibSp', 'Parch' ],
        'y_column': 'Survived',
        'n_estimators': 100
    },
    'train': {
        'x_columns': [ 'Pclass', 'Sex', 'SibSp', 'Parch' ]
    }
}""")
        for k, v in argument.items():
            if k not in stage.pipeline:
                continue
            stage.pipeline[k].update(v)

    def train(self, argument: dict = None):
        self.__update_func_kwargs(TRAIN, argument)
        self.solve([TRAIN])

    def inference(self, argument: dict = None):
        self.__update_func_kwargs(INFERENCE, argument)
        self.solve([INFERENCE])


class Standalone(OneTime):
    pass


class Sagemaker(OneTime):
    pass


class DaemonStatusType(Enum):
    """ 백그라운드 작업 상태 코드

    """
    WAITING = 'waiting'
    SETUP = 'setup'
    LOAD = 'load'
    RUN = 'run'
    SAVE = 'save'
    FAIL = 'fail'


class Daemon(Computing):
    """ 백그라운드에서 실행되며, redis를 통해 추론 데이터 수신시 작업 수행하는 데몬 클래스

    """
    CHANNEL_STATUS = "alo_status"
    CHANNEL_FAIL = "alo_fail"

    def __init__(self, **kwargs):
        self.__redis_status = None
        self.__redis_pubsub = None
        self.__status = None
        super().__init__(**kwargs)
        with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'config', 'redis_error_table.json'), 'r', encoding='utf-8') as file:
            self.redis_error_table = json.load(file)

    def init(self):
        super().init()
        if self.solution_metadata is None:
            raise AloErrors['ALO-INI-000']('solution_meta.yaml information is missing. "--system solution_meta.yaml option" must be specified.')
        try:
            self.__redis_status = self.get_redis(False,
                                                 self.solution_metadata.edgeapp_interface.redis_server_uri.host,
                                                 self.solution_metadata.edgeapp_interface.redis_server_uri.port,
                                                 self.solution_metadata.edgeapp_interface.redis_db_number)
            self.__redis_pubsub = self.get_redis(True,
                                                 self.solution_metadata.edgeapp_interface.redis_server_uri.host,
                                                 self.solution_metadata.edgeapp_interface.redis_server_uri.port,
                                                 self.solution_metadata.edgeapp_interface.redis_db_number)
            logger.debug("Redis Server(DB): %s(%s)",
                         self.solution_metadata.edgeapp_interface.redis_server_uri,
                         self.solution_metadata.edgeapp_interface.redis_db_number)
        except Exception as e:
            raise AloErrors['ALO-INI-004'](str(e)) from e
        self.pubsub_edgeapp('publish', self.CHANNEL_STATUS, "booting")

    def get_redis(self, is_strict, host, port, db_num):
        return (redis.StrictRedis if is_strict else redis.Redis)(host, port, db_num)

    def __send_edgeapp(self, redis, method, *args, **kwargs):
        try:
            result = getattr(redis, method)(*args, **kwargs)
            logger.debug("[REDIS] %s(%s, %s) : %s", method, args, kwargs, result)
            return result
        except Exception as e:
            raise AloErrors['ALO-INI-004'](f"{method}({args},{kwargs})") from e

    def interface_edgeapp(self, method, *args, **kwargs):
        return self.__send_edgeapp(self.__redis_status, method, *args, **kwargs)

    def pubsub_edgeapp(self, method, *args, **kwargs):
        if len(args) >= 2 and args[0] == self.CHANNEL_STATUS:
            self.__status = args[1]
        return self.__send_edgeapp(self.__redis_pubsub, method, *args, **kwargs)

    def error_to_code(self, error):
        code = "E000"
        comment = None
        if isinstance(error, AloError):
            # todo
            # error.code == ''
            comment = str(error)

        return {**self.redis_error_table[code], 'COMMENT': comment} if comment else self.redis_error_table[code]

    def solve(self):
        try:
            self.update_experimental_plan(INFERENCE)
            stage = getattr(self.experimental_plan.solution, INFERENCE)
            model_files = stage.get_model(settings.model_artifacts_path)
            logger.debug('[MODEL] List of imported model:\n%s', "\n".join(model_files))
            extract_file(model_files, settings.model_artifacts_path)
        except Exception as e:
            logger.exception(e)
            self.interface_edgeapp('rpush', 'inference_summary', json.dumps({'status': 'fail', 'message': str(e)}))
            self.interface_edgeapp('rpush', 'inference_artifacts', json.dumps({'status': 'fail', 'message': str(e)}))
            raise e

        logger.debug('[DAEMON] Get ready.')
        while True:
            try:
                self.pubsub_edgeapp('publish', self.CHANNEL_STATUS, DaemonStatusType.WAITING.value)
                response = self.interface_edgeapp('blpop', "request_inference", timeout=0)
                if response is None:
                    continue
                self.pubsub_edgeapp('publish', self.CHANNEL_STATUS, DaemonStatusType.SETUP.value)
                with Context() as context:
                    try:
                        self.solution_metadata = load_model(json.loads(response[1].decode('utf-8')).get('solution_metadata'), SolutionMetadata)
                        self.update_experimental_plan(INFERENCE)
                        self.pubsub_edgeapp('publish', self.CHANNEL_STATUS, DaemonStatusType.RUN.value)
                        self.exec_stage(context, INFERENCE)
                    except Exception as e:
                        self.artifact(context, INFERENCE)
                        raise e
                    else:
                        summary = json.dumps({'status': 'success', 'message': context.summary(INFERENCE)})
                        self.interface_edgeapp('rpush', 'inference_summary', summary)
                        self.pubsub_edgeapp('publish', self.CHANNEL_STATUS, DaemonStatusType.SAVE.value)
                        self.artifact(context, INFERENCE)
                        self.interface_edgeapp('rpush', 'inference_artifacts', summary)
            except Exception as e:
                logger.exception(e)
                logger.error('[DAEMON] Due to an error, the current step is skipped. Waiting for the next request')
                self.pubsub_edgeapp('publish', self.CHANNEL_STATUS, DaemonStatusType.FAIL.value)
                self.pubsub_edgeapp("publish", self.CHANNEL_FAIL, json.dumps(self.error_to_code(e)))
                self.interface_edgeapp('rpush', 'inference_summary', json.dumps({'status': 'fail', 'message': str(e)}))
                if self.__status == DaemonStatusType.SAVE.value:
                    self.interface_edgeapp('rpush', 'inference_artifacts', json.dumps({'status': 'fail', 'message': str(e)}))
                else:
                    self.interface_edgeapp('rpush', 'inference_summary', json.dumps({'status': 'fail', 'message': str(e)}))
                    self.interface_edgeapp('rpush', 'inference_artifacts', json.dumps({'status': 'fail', 'message': str(e)}))
                # backoff 적용시 현재 위치에 적용


alo_mode = {
    'local': Standalone,
    'standalone': Standalone,
    'sagemaker': Sagemaker,
    'loop': Daemon,
    'batch': Daemon,
    'daemon': Daemon,
}


def Alo():
    """ 실행 옵션에 따른 실행 방식 선택

    Returns: alo 객체

    """
    return alo_mode.get(settings.computing)()



import os
import sys
import argparse
import shutil

from alo.__version__ import __version__


def __run(args):
    from alo.alo import Alo
    from alo.model import settings, Git
    if args.name:
        settings.name = args.name
    if args.config:
        settings.config = args.config
    if args.system:
        settings.system = args.system
    if args.computing:
        settings.computing = args.computing
    settings.mode = None if args.mode == 'all' else args.mode
    if args.loop:
        settings.computing = 'daemon'
    if getattr(args, "git.url"):
        settings.git = Git(url=getattr(args, 'git.url'),
                           branch=getattr(args, 'git.branch') if getattr(args, 'git.branch') else 'main')
    if args.log_level:
        settings.log_level = args.log_level
    alo = Alo()
    alo.run()


def __template(args):
    # todo
    print("Coming soon.")


def __history(args):
    from alo.alo import Alo
    from alo.model import settings
    if args.config:
        settings.config = args.config
    alo = Alo()
    alo.history(type=args.mode, show_table=True, head=args.head, tail=args.tail)


def __register(args):
    import yaml
    import re

    def check_str_bytes(s, encoding='utf-8', bytes_limit = 5000):
            """ Check if string bytes is under 5000

            Args:
                s: string tobe checked
                encoding: method of string encoding(default: 'utf-8')

            Returns:
                True: bytes < 5000
                False: bytes >= 5000
            """
            byte_length = len(s.encode(encoding))
            if byte_length >= bytes_limit:
                raise ValueError(f"Input exceeds {bytes_limit} bytes limit (current: {byte_length} bytes)")
            return True

    def validate_name(name):
        if not name:  # Empty input is allowed
            return True

        if len(name) > 50:
            raise ValueError("Name must be 50 characters or less")

        # Check for Korean characters
        if any(ord(char) >= 0x3131 and ord(char) <= 0xD7A3 for char in name):
            raise ValueError("Name cannot contain Korean characters")

        # Only allow alphanumeric and hyphen
        if not re.match("^[a-zA-Z0-9-]*$", name):
            raise ValueError("Name can only contain letters, numbers, and hyphens")

        # Check for spaces
        if ' ' in name:
            raise ValueError("Name cannot contain spaces")

        return True

    def read_yaml(file_path):
        with open(file_path, 'r', encoding='utf-8') as file:
            data = yaml.safe_load(file)
        return data

    def write_yaml(data, file_path):
        with open(file_path, 'w', encoding='utf-8') as file:
            yaml.dump(data, file, default_flow_style=False, allow_unicode=True)

    def update_yaml(data, name=None, overview=None, detail=None):
        # Only update if the input is not empty
        if name and name.strip():  # name이 존재하고 공백이 아닌 경우에만 업데이트
            data['name'] = name
        if overview and overview.strip():  # overview가 존재하고 공백이 아닌 경우에만 업데이트
            data['overview'] = overview
        if detail:  # detail 리스트가 비어있지 않은 경우에만 업데이트
            data['detail'] = detail
        return data

    def copy_file_to_folder(src_file, dest_folder):
    # 복사하려는 파일이 존재하는지 확인합니다.
        if not os.path.isfile(src_file):
            print(f"{src_file} 파일을 찾을 수 없습니다.")
            return

        # 대상 폴더가 존재하지 않으면 생성합니다.
        if not os.path.exists(dest_folder):
            os.makedirs(dest_folder)

        # 파일명을 가져와 대상 폴더의 경로를 만듭니다.
        dest_file = os.path.join(dest_folder, os.path.basename(src_file))

        # 파일을 복사합니다.
        shutil.copy2(src_file, dest_file)
        print(f"{src_file} 파일이 {dest_file} 위치로 복사되었습니다.")

    from alo.solution_register import SolutionRegister
    src = os.getcwd()# os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'alo', 'example')
    settings = os.path.join(src, 'setting')
    solution_info = os.path.join(settings, 'solution_info.yaml')
    infra_config = os.path.join(settings, 'infra_config.yaml')

    data = read_yaml(solution_info)
    args.id = input("Please enter your AI Conductor ID: ")
    args.password = input("Please enter your AI Conductor password: ")
    # Get user inputs for updating
    name = input("Enter the new name (leave empty to keep current): ")
    validate_name(name)
    overview = input("Enter the new overview (leave empty to keep current): ")

    detail = []
    while True:
        add_detail = input("Do you want to add a detail? (If yes, type 'yes'; to skip, press enter): ").strip().lower()
        if add_detail == 'yes':
            content = input("Enter the content for the detail: ")
            check_str_bytes(content)
            title = input("Enter the title for the detail: ")
            check_str_bytes(title)
            detail.append({"content": content, "title": title})
        elif add_detail =='' :
            break
        else :
            raise ValueError("Invalid input! You must type 'yes' or press enter to skip.")
    data = update_yaml(data, name, overview, detail)
    write_yaml(data, solution_info)

    current_settings_dir = os.path.join(os.getcwd(), 'setting')
    os.makedirs(current_settings_dir, exist_ok=True)
    # copy_file_to_folder(solution_info, current_settings_dir)
    # copy_file_to_folder(infra_config, current_settings_dir)

    solution_register = SolutionRegister(args.id, args.password)
    solution_register.register()


def __update(args):
    from alo.solution_register import SolutionRegister
    solution_register = SolutionRegister(args.id, args.password)
    solution_register.update()


def __delete(args):
    from alo.solution_register import SolutionRegister
    solution_register = SolutionRegister(args.id, args.password)
    solution_register.delete()


def __example(args):
    src = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'example', args.name)
    for item in os.listdir(src):
        s = os.path.join(src, item)
        d = os.path.join(os.getcwd(), item)
        if os.path.isdir(s):
            shutil.copytree(s, d)
        else:
            shutil.copy2(s, d)

    print(f"A {args.name} template file has been created in the current path.")
    print("Run alo")

def __docker(args):

    def print_ubuntu_package_instructions():

        instructions = """
        ### 우분투 패키지 설치 안내문 🌈

        안녕하세요! Dockerfile을 통해 우분투 패키지를 설치해보겠습니다. 아래의 단계를 따라 패키지를 추가해보세요:

        1. Dockerfile에서 `apt-get update` 명령을 포함하여 APT 패키지 목록을 업데이트하세요.

        2. `apt-get install` 명령을 사용하여 패키지를 설치하세요. `--no-install-recommends` 옵션을 사용하면 불필요한 의존성을 최소화할 수 있습니다.

        **예제**:
        우분투 패키지 `curl`을 설치하고 싶다면, Dockerfile의 해당 부분에 다음과 같이 추가하세요:

        ```dockerfile
        RUN apt-get update && \\
            apt-get install -y --no-install-recommends \\
            curl \\
            && rm -rf /var/lib/apt/lists/*

        추가된 curl 패키지는 Docker 컨테이너 내에서 사용 가능합니다.
        즐거운 ALO 생활 되세요 🐧
        """

        print(instructions)

    def print_cuda_instructions():

        instructions = """
        ### alo docker --gpu 를 실행한 GPU용 Dockerfile 작성자 용
        ### Docker container cuda와 cudnn 설정에 관한 안내문 🌈

        - CUDA 버전 및 CuDNN 버전을 환경 변수로 정의합니다. tensorflow, torch 버전에 따라
        호환되는 CUDA_VER 및 CUDNN_VER 버전을 작성합니다.
        << 작성 예시 >>
        ## torch >= 2.1.0
        FROM nvidia/cuda:11.8.0-devel-ubuntu22.04
        ARG CUDA_VER=11.8
        ############################################################
        ##  torch <= 2.0.1
        FROM nvidia/cuda:11.7.1-devel-ubuntu22.04
        ARG CUDA_VER=11.7
        ############################################################
        ## tensorflow 2.15
        FROM nvidia/cuda:12.2.2-cudnn8-devel-ubuntu22.04
        ARG CUDA_VER=12.2
        ############################################################
        ## tensorflow 2.14
        FROM nvidia/cuda:11.8.0-devel-ubuntu22.04
        ARG CUDA_VER=11.8
        ARG CUDNN_VER=8.7.0
        ############################################################
        ## tensorflow 2.12 ~ 2.13
        FROM nvidia/cuda:11.8.0-devel-ubuntu22.04
        ARG CUDA_VER=11.8
        ARG CUDNN_VER=8.6.0
        ############################################################

        가령, torch나 tensorflow 2.15 버전 이상부터는 CUDNN_VER은 미작성합니다.

        - 참고 사항
        CUDA 및 CuDNN 설치는 CUDA 버전과 CuDNN의 호환성을 반드시 확인해야 합니다.
        NVIDIA 사이트에서 버전별 설치 가이드를 참고하면 더욱 정확한 설치가 가능합니다.
        주의사항: 호환성을 잘못 맞추면 예상치 못한 에러가 발생할 수 있습니다.
        도움이 되셨길 바랍니다! 필요에 따라 Dockerfile을 수정하여 나만의 Docker 이미지를 만들어보세요. 🚀 """

        print(instructions)

    subdir = 'dockerfile_gpu' if args.gpu else 'dockerfile_cpu'
    dockerfile_src = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'Dockerfiles', 'register', subdir, 'Dockerfile')
    dockerfile_dest = os.path.join(os.getcwd(), 'Dockerfile')
    print_ubuntu_package_instructions()
    print_cuda_instructions()
    if os.path.exists(dockerfile_src):
        shutil.copy2(dockerfile_src, dockerfile_dest)
        print(f"Dockerfile has been copied to the current path.")
    else:
        print("Error: Dockerfile not found.")

def main():
    if len(sys.argv) > 1:
        if sys.argv[-1] in ['-v', '--version']:
            print(__version__)
            return
        if sys.argv[1] in ['-h', '--help']:
            pass
        elif sys.argv[1] not in ['run', 'history', 'register', 'update', 'delete', 'template', 'example', 'docker']:  # v1 호환
            sys.argv.insert(1, 'run')
    else:
        sys.argv.insert(1, 'run')

    parser = argparse.ArgumentParser('alo', description='ALO(AI Learning Organizer)')
    parser.add_argument('-v', '--version', action='version', version=__version__)
    subparsers = parser.add_subparsers(dest='command')

    cmd_exec = subparsers.add_parser('run', description='Run alo')
    cmd_exec.add_argument('--name', type=str, help='name of solution')
    cmd_exec.add_argument('--mode', type=str, default='all', choices=['train', 'inference', 'all'], help='ALO mode: train, inference, all')
    cmd_exec.add_argument("--loop", dest='loop', action='store_true', help="On/off infinite loop: True, False")
    cmd_exec.add_argument("--computing", type=str, default="local", choices=['local', 'daemon'], help="training resource: local, ...")
    cmd_exec.add_argument('--config', type=str, help='path of experimental_plan.yaml')
    cmd_exec.add_argument('--system', type=str, help='path of solution_metadata.yaml')
    cmd_exec.add_argument('--git.url', type=str, help='url of git repository')
    cmd_exec.add_argument('--git.branch', type=str, help='branch name of git repository')
    cmd_exec.add_argument('--log_level', type=str, default="DEBUG", choices=['NOTSET', 'DEBUG', 'INFO', 'WARNING', 'ERROR'], help='log level')

    cmd_history = subparsers.add_parser('history', description='Run history')
    cmd_history.add_argument('--config', type=str, help='path of experimental_plan.yaml')
    cmd_history.add_argument('--mode', default=['train', 'inference'], choices=['train', 'inference'], nargs='+', help='train, inference')
    cmd_history.add_argument("--head", type=int, default=None, help="output the last part of history")
    cmd_history.add_argument("--tail", type=int, default=None, help="output the first part of history")

    cmd_template = subparsers.add_parser('template', description='Create titanic template')

    cmd_register = subparsers.add_parser('register', description='Create new solution')
    cmd_register.add_argument('--id', help='user id of AI conductor')
    cmd_register.add_argument('--password', help='user password of AI conductor')
    cmd_register.add_argument('--description', default=None, help='description')

    cmd_update = subparsers.add_parser('update', description='Update a solution')
    cmd_update.add_argument('--id', required=True, help='user id of AI conductor')
    cmd_update.add_argument('--password', required=True, help='user password of AI conductor')

    cmd_delete = subparsers.add_parser('delete', description='Delete a solution')
    cmd_delete.add_argument('--id', required=True, help='user id of AI conductor')
    cmd_delete.add_argument('--password', required=True, help='user password of AI conductor')

    cmd_example = subparsers.add_parser('example', description='Create ALO example')
    cmd_example.add_argument('--name', default='titanic', choices=['titanic'], help='Example of ALO')

    # Add docker command parser
    cmd_docker = subparsers.add_parser('docker', description='Create Dockerfile for ALO')
    cmd_docker.add_argument('--gpu', action='store_true', help='Provide GPU Dockerfile sample')

    args = parser.parse_args()

    commands = {'run': __run,
                'template': __template,
                'history': __history,
                'register': __register,
                'update': __update,
                'delete': __delete,
                'example': __example,
                'docker': __docker,
                }
    commands[args.command](args)
